#ifndef E_ASET_H
#define E_ASET_H
#include <stdio.h>
#include <stdarg.h>

extern e_obj_descriptor e_obj_descriptor_aset;
#define E_IS_ASET(P) (E_OBJ_DESCR(P) == &e_obj_descriptor_aset)
#define E_TYPE_ASET (&e_obj_descriptor_aset)

extern e_obj_descriptor e_obj_descriptor_aset_item;
#define E_IS_ASET_ITEM(P) (E_OBJ_DESCR(P) == &e_obj_descriptor_aset_item)
#define E_TYPE_ASET_ITEM (&e_obj_descriptor_aset_item)


#define E_ASET_ADD_MODE_UNIQUE_VAL 1
#define E_ASET_ADD_MODE_UNIQUE_KEY (1<<1)
#define E_ASET_ADD_MODE_NO_REPLACE (1<<2)

                  /* Property modes */
#define E_ASET_PROPERTY_ADD_SET 1
#define E_ASET_PROPERTY_MULTI   2

typedef struct e_AsItem {
  e_obj h;           /* object header                   */
  struct e_AsItem *next; /* Next item                   */
  struct e_AsItem *prev; /* Previous item object        */
  e_name *name;          /* Item name                   */
  void *val;             /* item value                  */
} e_aset_item;


typedef struct {
  e_obj h;              /* object header                */
  e_aset_item *first;   /* First item                   */
  e_aset_item *last;    /* Last item                    */
  e_aset_item *cur;     /* Current item                 */
  int cardinal;         /* Number of item in the aset   */
} e_aset;

/*****************************  ASET **********************************/
e_aset *e_aset_new(e_aset *asp); /* [0;+1]  */
e_aset_item *e_aset_item_new(e_aset_item *aip,e_name *np,void *val);
                                                       /* [0,-1,-1;+1] */
void e_aset_free(void *asp);      /* [-1;.] */
void e_aset_item_free(void *aip); /* [-1;.] */
void *e_aset_to_string(void *asp,int tab,e_flag mode); /* [0,.,.;+1] */
void *e_aset_item_to_string(void *aip,int tab,e_flag mode); /* [0,.,.;+1] */
void *e_aset_clone(void *asp,e_flag mode);      /* [0,.;+1] */
void *e_aset_item_clone(void *aip,e_flag mode); /* [0,.;+1] */
int e_aset_equals(void *p1,void *p2);           /* [0,0;.] */
int e_aset_item_equals(void *p1,void *p2);      /* [0,0;.] */
int e_aset_next(void *asp,int mode,void **obj); /* [0,*+1;.] */

/************************ ADD TO ASET *************************************/
e_aset *e_aset_append_item(e_aset *asp,e_name *np, void *val); 
                                                /* [-1,-1,-1;+1] */
e_aset *e_aset_prepend_item(e_aset *asp,e_name *np, void *val);
                                                /* [-1,-1,-1;+1] */
e_aset *e_aset_add_item(e_aset *asp,e_name *np, void *val,int mode);
                                                /* [-1,-1,-1,.;+1] */
e_aset *e_aset_add_aset(e_aset *asp,e_aset *aspv,int mode);
                                                /* [-1,-1,.;+1] */
e_aset *e_aset_append(e_aset *asp,e_aset *asp2);/* [-1,-1;+1] */
e_aset *e_aset_fill(e_aset *asp,e_aset *asp2);  /* [-1,-1;+1] */

/************************ MODIFY ASET* *************************************/
e_aset_item *e_aset_scan_by_key(e_aset *asp, e_name *key,int mode);
                                                          /* [0,-1,.;0]   */ 
int e_aset_get_item_mode_by_key(e_aset *asp, e_name *key,int n);
                                                          /* [0,-1;.]     */
int e_aset_get_item_mode_by_index(e_aset *asp, int index);/* [0,.;.]      */
int e_aset_set_item_mode_by_key(e_aset *asp, e_name *key,
                                       int mode,int op,int n); 
                                                          /* [0,-1,.,.,.;.] */
int e_aset_set_item_mode_by_index(e_aset *asp, int index,int mode,int op); 
                                                          /* [0,.,.,.;.]  */
int e_aset_set_item_value(e_aset_item *aip,void *val);    /* [0,-1:.]     */
int e_aset_set_value_by_key(e_aset *asp,e_name *key, void *val,int n);  
                                                          /* [0,-1,-1,.;.]*/
int e_aset_set_value_by_index(e_aset *asp,int index, void *val);  
                                                          /* [0,.,-1;.]   */
e_aset *e_aset_apply_fct(e_aset *asp,void *p,int i,
                e_obj *(* fct)(e_obj *op,void *p,int i)); /* [-1,.,.,.;+1] */

/************************ EXTRACT FROM ASET ********************************/
int e_aset_get_key_index(e_aset *asp, e_name *np,int start);   /* [0,-1,.;.] */
int e_aset_get_value_index(e_aset *asp,void *obj,int start); /* [0,-1,.;.] */
int e_aset_get_item_index(e_aset *asp, e_name *np,void *val,int start);
                                                /* [0,-1,-1,.;.] */
void *e_aset_get_value_by_key(e_aset *asp, e_name *np,int start);
                                                /* [0,-1,.;+1] */
void *e_aset_get_value_by_index(e_aset *asp,int index);       /* [-0,.;+1] */
e_aset_item *e_aset_get_item_by_key(e_aset *asp, e_name *np,int n); 
                                                              /* [0,-1,.;0] */
e_aset_item *e_aset_get_item_by_value(e_aset *asp, void *val);/* [0,-1;0] */
e_aset_item *e_aset_get_item_by_index(e_aset *asp,int index); /* [0,.;0]  */
e_aset *e_aset_get_keys(e_aset *asp);                         /* [0;+1] */
e_aset *e_aset_anonymous_subset(e_aset *asp);                 /* [0;+1] */
e_aset *e_aset_named_subset(e_aset *asp);                     /* [0;+1] */
e_aset *e_aset_subset(e_aset *asp,int from,int to);           /* [0,.,.;+1] */
e_aset *e_aset_subset_by_key_match(e_aset *asp,e_string *match);
                                                              /* [0,-1;+1] */
e_aset *e_aset_subset_by_test(e_aset *asp,void *val,
                          int (*test)(e_obj *op,void *val));  /* [0,0,0;+1] */
char *e_aset_get_string_pointer_by_key(e_aset *asp,e_name *np); /* [0,0,-] */
void *e_aset_get_number_pointer_by_key(e_aset *asp,e_name *np,int type);
                                                              /* [0,0,-;-] */

/***************************** MODIFY ASET **********************************/
e_aset *e_aset_sort_by_key(e_aset *asp,e_flag mode);      /* [-1,.;+1] */
e_aset *e_aset_sort_by_value(e_aset *asp,e_flag mode,
                   int (* testf)(void *,void *,int));     /* [-1,.;+1] */
int e_aset_remove_item_by_index(e_aset *asp,int index);   /* [0,.;.]   */
int e_aset_remove_cur_item(e_aset *asp);                  /* [0;.]     */
int e_aset_remove_item_by_key(e_aset *asp,e_name *key,int index);
                                                          /* [0,0,.;.] */
e_aset *e_aset_diff_by_key(e_aset *asp1,e_aset *asp2, int max);    
                                                          /* [-1,0,.;+1] */
e_aset *e_aset_diff_by_value(e_aset *asp1,e_aset *asp2, int max); 
                                                          /* [-1,0,.;+1] */
e_aset *e_aset_common_by_key(e_aset *asp1,e_aset *asp2);  /* [-1,0:+1] */
e_aset *e_aset_common_by_value(e_aset *asp1,e_aset *asp2);/* [-1,0:+1] */
int e_aset_is_in(e_aset *asp,e_name *key,void *val);      /* [0,0,0;.]   */
                                                         

/************************ OTHERS ********************************/
e_aset *e_aset_split(e_string *sp,char *sep, char *psep);     /* [-1,.,.;+1] */
e_aset *e_aset_add_property(e_aset *asp, e_name *key, void *val,int mode);
                                                          /* [-1,-1,-1,.;+1] */
e_aset *e_aset_get_properties(e_aset *asp, void *val);
							  /* [-1,-1;+1]*/
e_aset *make_aset(void *ob1, ...);                    /* [-1,-1, ... ,-1;+1]*/
e_aset *make_named_aset(void *n1, ...);               /* [-1,-1, ... ,-1;+1]*/

#define E_ASET_FIRST(P) (((e_aset *)P)->first)
#define E_ASET_LAST(P) (((e_aset *)P)->last)
#define E_ASET_CUR(P) (((e_aset *)P)->cur)
#define E_ASET_CARDINAL(P) (((e_aset *)P)->cardinal)

#define E_ASET_ITEM_PREV(P) (((e_aset_item *)P)->prev)
#define E_ASET_ITEM_NEXT(P) (((e_aset_item *)P)->next)
#define E_ASET_ITEM_NAME(P) (((e_aset_item *)P)->name)
#define E_ASET_ITEM_VAL(P) (((e_aset_item *)P)->val)

#endif
