#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <md5.h>

#define HASH_Q2 1048573
#define HASH_Q1 2039
#define READ_BUFF_SIZE 1024
/*
#define READ_BUFF_SIZE 32768
*/
unsigned int n,m,r,s;

unsigned int hash_file(char *file_name);
unsigned int block_hash_file(char *file_name);
unsigned int md5_hash(char *file_name,char *buff);
void MD5_2_String(unsigned char *digest,char *buff);

int main(int argc, char **argv)
{ int i,md5 = 0;
  char check_sum[40];

  if((argc < 2) || !strcmp(argv[1],"-help")  || !strcmp(argv[1],"-h"))
  { printf("SYNTAX : hashfile [-md5] FILENAMES");
    exit(1);
  }

  i = 1;
  while(i < argc)
  { if(!strcmp(argv[i],"-md5")) md5 = 1;
    else
    { if(md5)
      { md5_hash(argv[i],check_sum);
        printf("%s %s\n",check_sum,argv[i]);
      }
      else printf("%u %s\n",block_hash_file(argv[i]),argv[i]);
    }
    i++;
  }

  return(0);
}


/****************************************/
/* compute a hash value for a file      */
/* file_name : name of the file to hash */
/* Return : hash code 0 <= HASH_Q2<     */
/*          0xffffffff : error          */
/****************************************/
unsigned int block_hash_file(char *file_name)
{ FILE *fpt;
  int c,i,size;
  unsigned int n,m,r,s;
  char buff[READ_BUFF_SIZE];

  if(!(fpt = fopen(file_name,"r")))
  { printf("hash_file error : can't open file %s\n",file_name);
    return(0xffffffff);
  }

  m=r=1;

  while((size=fread(buff,1,READ_BUFF_SIZE,fpt))>0)
  { i = 0;
    while(i < size)
    { c = buff[i++];
      if(c < 0) c = -c+128;

      if((c > 127) || ((c < 32) && (c != '\t') && (c!= '\n')  && (c!= '\r')))
      { s = r >> 1;
      }     
      else s = r;

      n = c+1;
      m = (((m*n)%HASH_Q1)+s)%HASH_Q1;
      r = ((m+1)*(r+1))%HASH_Q2;
    }
  }
  fclose(fpt);
  return(r);
}

/****************************************/
/* compute the md5 hash value for a file*/
/* file_name : name of the file to hash */
/* Return the md5 value in buff         */
/* Return : 1 : OK                      */
/*          0 : error                   */
/****************************************/
unsigned int md5_hash(char *file_name,char *buff)
{ FILE *fpt;
  MD5_CTX context;
  int len;
  unsigned char buffer[READ_BUFF_SIZE], digest[16];

  if(!(fpt = fopen(file_name,"r")))
  { printf("hash_file error : can't open file %s\n",file_name);
    return(0);
  }

  MD5Init (&context);
  while ((len = fread (buffer, 1, READ_BUFF_SIZE, fpt)))
  { MD5Update (&context, buffer, len);
  }
  MD5Final (digest, &context);

  fclose (fpt);
  MD5_2_String(digest,buff);

  return(1);
}

/**************************************************/
/* Convert the md5 value to an hexadecinal string */
/* digest : array to convert.                     */
/* buff : destination for the string.             */
/**************************************************/
void MD5_2_String(unsigned char *digest,char *buff)
{
  unsigned int i,j;

  j = 0;
  for (i = 0; i < 16; i++)
  { sprintf(&buff[j],"%02x", digest[i]);
    j += 2;
  }
}
