#include <easy.h>
#include <stdio.h>
#define __USE_BSD
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/unistd.h>
#include <fcntl.h>
#include <unistd.h>
#include "sysf.h"
#include "sysf_com.h"
#include "sysf_init.h"
#include "sysf_dir.h"
#include <pwd.h>
#include <grp.h>
#undef __USE_BSD

extern char HostName[E_LINE_SIZE];
extern sysf_cf sysf_cf_vars;

/************************************************/
/* Display info about the selected config file  */
/* name : config name                           */
/* sip : pointer to sysfile_file info structure */
/************************************************/
int get_info(char *name,sysfile_info *sip,int echo)
{ char sys_time[E_LINE_SIZE],conf_time[E_LINE_SIZE];
  char uid_name[10],gid_group[10];
  char *type;

  strftime(sys_time,E_LINE_SIZE,"%H:%M:%S %d/%m/%Y",
           localtime(&sip->sys_file.time));
  strftime(conf_time,E_LINE_SIZE,"%H:%M:%S %d/%m/%Y",
           localtime(&sip->config_file.time));


  if(!sip->sys_file.exists)
  { mprintf(" System file (%s) does not exist.\n",sip->sys_file.FileName);
  }
  else if(sip->flags & E_SYSF_FLAG_g)
  { mprintf(" System DIR LIST (%s) : %s\n",
               sys_time,sip->sys_file.FileName);
  }
  else if(IS_SYMLINK(sip->sys_file.type))
  { mprintf(" System file (%s) is a SYMBOLIC LINK pointing to\n      %s\n",
	    sip->sys_file.FileName,sip->sys_file.LinkPath);
  }
  else 
  { type = "file";
    if(IS_DIR(sip->sys_file.type)) type = "DIR";

    get_user_name(sip->sys_file.uid,uid_name);
    get_group_name(sip->sys_file.gid,gid_group);
    mprintf(" System %s (%s) : %s (%d)\n",
               type,sys_time,sip->sys_file.FileName,sip->sys_file.size);
    mprintf("             uid=%-8.8s   gid=%-8.8s mode=%o\n",
                          uid_name,gid_group,sip->sys_file.mode);
  }

  if(!sip->config_file.exists)
  { mprintf(" Missing config file for %s.\n",name);
  }
  else if(sip->config_file.none)
  { mprintf(" Config file (%s) : %s\n             SET TO NONE\n",
                conf_time,sip->config_file.FileName);
  }
  else if(sip->flags & E_SYSF_FLAG_g)
  { mprintf(" Config DIR LIST (%s) : %s\n",
               conf_time,sip->config_file.FileName);
  }
  else if(IS_SYMLINK(sip->config_file.type))
  { mprintf(
   " Config file (%s) \n   defines a SYMBOLIC LINK pointing to\n      %s\n",
	    sip->config_file.FileName,sip->config_file.LinkPath);
  }
  else
  { get_user_name(sip->config_file.uid,uid_name);
    get_group_name(sip->config_file.gid,gid_group);

    if(IS_DIR(sip->config_file.type))
    { mprintf(
   " Config file (%s) : %s \n   defines a DIRECTORY:\n",
	    conf_time,sip->config_file.FileName);
    }
    else
    { mprintf(" Config file (%s) : %s (%d)\n",conf_time,
                sip->config_file.FileName,sip->config_file.size);
    }

    mprintf("             uid=%-8.8s   gid=%-8.8s mode=%o\n",uid_name,
               gid_group, sip->config_file.mode);

  }

  if(!sip->sys_file.exists)
  { if(!sip->config_file.none)
    { mprintf(" Missing system file for %s\n",name);
    }
  }

  if(sip->config_file.none && sip->sys_file.exists)
  { mprintf("System file %s exists while \nConfig file %s is set to NONE\n",
             sip->sys_file.FileName,sip->config_file.FileName);
    goto end;
  }
  
  if(!test_file_type(sip->flags,&sip->sys_file,&sip->config_file,0))
  { goto end;
  }

  if(sip->flags & E_SYSF_FLAG_l)
  { if(sip->differ)
    { if(sip->sys_file.exists && sip->config_file.exists)
      { mprintf(
   " System file is a SYMBOLIC LINK different from the LINK defined
     by the config file!\n");
      }
      goto end;
    }

    if(sip->sys_file.exists && sip->config_file.exists)
      mprintf(" The System file and the Config file are identical.\n");
    goto end;
  }

  if(sip->flags & E_SYSF_FLAG_d) /* directories */
  { if(!sip->differ) /* differences have been covered above! */
    { mprintf(
         " The System and the Config directories are up to date.\n");
    }
    goto end;
  }
  else
  { if(!sip->config_file.exists || !sip->sys_file.exists) goto end;
 
    if(sip->differ)
    { if(sip->sys_file_older)
      { mprintf("   %s%s\n   %s\n",
                "The System file and the Config file differ ",
                "and the System file is the oldest.",
                "The system file needs to be updated."
               );
      }
      else
      { mprintf("   %s%s\n   %s\n",
                "The System file and the Config file differ ",
                "and the Config file is the oldest.",
                "The system file needs to be archived."
               );
      }
    }
    else
    { if(sip->sys_file_older)
      { mprintf("   %s%s\n",
                "The System file and the Config file are identical ",
                "but the System file is the oldest."
               );
      }
      else
      { mprintf("   %s\n",
                "The System file and the Config file are identical."
               );
      }
    }
  }

  /* compare the attributes */
  if(sip->mode_differ)
  { if (sip->sys_file.mode != sip->config_file.mode)
    { mprintf("  File mode differ : %o %o\n",
  	         sip->sys_file.mode,sip->config_file.mode);
    }
    if (sip->sys_file.uid != sip->config_file.uid)
    { mprintf("  File uid differ : %d %d\n",
  	         sip->sys_file.uid,sip->config_file.uid);
    }
    if (sip->sys_file.gid != sip->config_file.gid)
    { mprintf("  File gid differ : %d %d\n",
  	         sip->sys_file.gid,sip->config_file.gid);
    }
  }

end:
  mprintf("\n");
  return(1);
}

/*****************************************************************/
/* Check that the config file and the system file are identical. */
/* Display a message only if the two file differ.                */
/* Increment the Check counter when the files are differents.    */
/* name : config name                                            */
/* sip : pointer to sysfile_file info structure                  */
/*****************************************************************/
int check(char *name,sysfile_info *sip,int echo)
{ char sys_time[E_LINE_SIZE],conf_time[E_LINE_SIZE];
  char uid_name_s[10],gid_group_s[10],uid_name_c[10],gid_group_c[10];

  if(!sip->sys_file.exists)
  { if(echo || !sip->config_file.none) 
    { mprintf("%s:\n",name);
      mprintf("System file %s does not exist.\n",sip->sys_file.FileName);
    }

    if(!sip->config_file.exists)
      mprintf("No config file for %s\n",name);

    return(1);
  }

  if(!sip->config_file.exists)
  { mprintf("No config file for %s\n",name);
    return(1);
  }

  if(sip->config_file.none && sip->sys_file.exists)
  { if(!(sip->flags & E_SYSF_FLAG_g))
    { mprintf("%s:\n",name);
      mprintf("System file %s exists while\nConfig file %s is set to NONE\n",
               sip->sys_file.FileName,sip->config_file.FileName);
    }
    return(1);
  }

  if(!test_file_type(sip->flags,&sip->sys_file,&sip->config_file,0))
  { return(1);
  }

  if(sip->flags & E_SYSF_FLAG_l)
  { if(sip->differ)
    { mprintf("%s:\n",name);
      mprintf(" System file (%s)\n  is a SYMBOLIC LINK to: ->%s<-
 which DIFFERS from the LINK defined by the config file\n  (%s) : ->%s<-\n",
     sip->sys_file.FileName,sip->sys_file.LinkPath,
     sip->config_file.FileName,sip->config_file.LinkPath);
      return(1);
    }
    return(1);
  }

  if(sip->differ || sip->mode_differ)
  { strftime(sys_time,E_LINE_SIZE,"%H:%M:%S %d/%m/%Y",
           localtime(&sip->sys_file.time));
    strftime(conf_time,E_LINE_SIZE,"%H:%M:%S %d/%m/%Y",
           localtime(&sip->config_file.time));
    get_user_name(sip->sys_file.uid,uid_name_s);
    get_group_name(sip->sys_file.gid,gid_group_s);
    get_user_name(sip->config_file.uid,uid_name_c);
    get_group_name(sip->config_file.gid,gid_group_c);

    sysfile.CheckCounter++;

    if(sip->flags & E_SYSF_FLAG_g)
    { if(sip->differ)
      { mprintf("%s:\n",name);
        mprintf(" System and Config DIR LIST differ:\n");
    
        mprintf(" System DIR LIST (%s) : %s\n",sys_time,
                                               sip->sys_file.FileName);
         mprintf(" Config DIR LIST (%s) : %s\n",conf_time,
                                                sip->config_file.FileName);
         dir_check(sip->diff_asp,echo);
      }
      if(sip->sys_file_older)
      { mprintf(" The Config DIR LIST was last modified.\n\n");
      }
      else
      { mprintf(" The System DIR LIST was last modified.\n\n");
      }
      return(1);
    }

    if(sip->flags & E_SYSF_FLAG_d)
    { if(sip->differ)
      { mprintf("%s:\n",name);
        mprintf(" System DIRECTORY %s is not up to date!\n",
                sip->sys_file.FileName);
      }
      else if (sip->mode_differ)
      { mprintf("%s:\n",name);
        mprintf(" System and Config DIRECTORY attributes differ:\n");
    
        mprintf(" System DIR (%s) : %s\n",sys_time,sip->sys_file.FileName);
        mprintf("             uid=%-8.8s   gid=%-8.8s mode=%o\n",
                              uid_name_s,gid_group_s,sip->sys_file.mode);

        mprintf(" Config DIR (%s) : %s\n",conf_time,sip->config_file.FileName);
        mprintf("             uid=%-8.8s   gid=%-8.8s mode=%o\n",
                              uid_name_c,gid_group_c,sip->config_file.mode);
      }
      return(1);
    }


    mprintf("%s:\n",name);
    mprintf(" System and Config file Differ:\n");
    
    mprintf(" System file (%s) : %s (%d)\n",sys_time,sip->sys_file.FileName,
                           sip->sys_file.size);
    mprintf("              uid=%-8.8s   gid=%-8.8s mode=%o\n",
                           uid_name_s,gid_group_s,sip->sys_file.mode);

    mprintf(" Config file (%s) : %s (%d)\n",conf_time,
                           sip->config_file.FileName,sip->config_file.size);
    mprintf("              uid=%-8.8s   gid=%-8.8s mode=%o\n",
                           uid_name_c,gid_group_c,sip->config_file.mode); 

    if(sip->sys_file_older)
    { mprintf(" The System file is the oldest.\n\n");
    }
    else
    { mprintf(" The Config file is the oldest.\n\n");
    }    
  }
  return(1);
}

/*******************************************************************/
/* List the config entries for which either the config file or the */
/* system file does not exist.                                     */
/* If config file is set to none, non existing system files are not*/
/* listed                                                          */
/* If neither of them exist, nothing is displayed                  */
/* name : config name                                              */
/* sip : pointer to sysfile_file info structure                    */
/*******************************************************************/
int is_not_conf(char *name,sysfile_info *sip,int echo)
{ 

  if(sip->config_file.none) 
  { /* NONE : config file must not exist or be a dir list */
    if(!sip->sys_file.exists || (sip->flags & E_SYSF_FLAG_g)) return(1);
  }

  if(sip->config_file.exists && sip->sys_file.exists) return(1);

/*  
  if(!sip->config_file.exists && !sip->sys_file.exists)
  { return(1);
  }
*/

  mprintf("%s :\n",name);

  if(!sip->sys_file.exists)
  { mprintf("System file %s does not exist.\n",sip->sys_file.FileName);
  }

  if(!sip->config_file.exists)
  { mprintf("No config file for %s\n",name);
  }

  return(1);
}

/******************************************************************/
/* Copy the config file into the system file.                     */
/* The config file must already exist.                            */
/* they are the config file for this config name.                 */
/* name : config name                                             */
/* sip : pointer to sysfile_file info structure                   */
/******************************************************************/
int install(char *name,sysfile_info *sip,int echo)
{ 
  if(!sip->config_file.exists)
  { mprintf("ERROR: No config file for %s\n",name);
    return(1);
  }

  if(sip->config_file.none)
  { /* remove the system file if it exists */
    if(!sip->sys_file.exists || (sip->flags & E_SYSF_FLAG_g)) 
    { /* system does not exists or dir list -> nothing to do */
      goto end;
    }
     /* directories are removed only if they are empty. */
    if (sip->flags & E_SYSF_FLAG_d)
    { E_NULL_ERROR(remove_dir(sip->sys_file.FileName,echo));
    }
    else
    { E_NULL_ERROR(remove_file(sip->sys_file.FileName,echo));
    }
  }
  else if(!test_file_type(sip->flags,0,&sip->config_file,0))
  { return(1);
  }
  else if (sip->flags & E_SYSF_FLAG_l)
  { /* create or update the symbolic link */
    if(!sip->sys_file.exists)
    { mprintf("Creating the system file (symbolic link) %s\n",
               sip->sys_file.FileName);
    }

    E_NULL_ERROR(remove_file(sip->sys_file.FileName,echo));

    E_NULL_ERROR(make_sym_link(sip->config_file.LinkPath,
                                    sip->sys_file.FileName,echo));
  }
  else if (sip->flags & E_SYSF_FLAG_d)
  {           /* create the directory recursively */
    if(!sip->sys_file.exists) 
    { mprintf("Creating the system directory %s\n",sip->sys_file.FileName);
      if(!mk_path_dir(sip->sys_file.FileName,echo)) goto error;
      sip->sys_file.exists = 1;
    }
    install_file_props(sip,1,1,1,echo);
  }
  else if (sip->flags & E_SYSF_FLAG_g)
  {           /* create the directory recursively */
    if(!sip->sys_file.exists)
    { mprintf("Creating the system directory (DIR LIST) %s\n",
               sip->sys_file.FileName);
      if(!mk_path_dir(sip->sys_file.FileName,echo)) goto error;
      sip->sys_file.exists = 1;

              /* we now reconfig the DIR LIST */
      E_NULL_ERROR(sip->dir_sys_asp = 
           dir_read_sys(sip->sys_file.FileName,sip->dir_config_asp));

      if (sip->dir_sys_asp && sip->dir_config_asp)
      { sip->diff_asp = dir_list_diff(sip->dir_sys_asp,sip->dir_config_asp);
      }
      else sip->diff_asp = 0;
    }

              /* install the dir list */
    if(!dir_install(sip,sip->diff_asp,sip->sys_file.FileName,echo,0))
    { mprintf("Unable to fully install the DIR LIST  %s\n",
              sip->sys_file.FileName);
    }
  }
  else
  { if(!sip->sys_file.exists)
    { mprintf("Creating the system file %s\n",sip->sys_file.FileName);
    }

    E_NULL_ERROR(copy_file(sip->config_file.FileName,
                                 sip->sys_file.FileName,echo));
  }

end:
  return(1);

error:
  return(0);
}

/*********************************************************************/
/* Copy the config file into the system file.                        */
/* The config file must already exist.                               */
/* Make a copy of the system file with the .DIST suffix if it        */
/* does not exists yet.                                              */
/* Do not create .DIST file if flags r or d (directory) is set!      */
/* name : config name                                                */
/* sip : pointer to sysfile_file info structure                      */
/*********************************************************************/
int first_install(char *name,sysfile_info *sip,int echo)
{ e_string *sp = 0;

  if(!sip->config_file.exists)
  { mprintf("ERROR: The config file %s does not exist!\n",
            sip->config_file.FileName);
    return(1);
  }

  if(!test_file_type(sip->flags,0,&sip->config_file,0))
  { return(1);
  }

  if(sip->sys_file.exists && !(sip->flags & E_SYSF_FLAG_r)
     && !(sip->flags & E_SYSF_FLAG_g) && !(sip->flags & E_SYSF_FLAG_d))
  { /* create the file CONFIG.DIST if it does not exist */
    E_NULL_ERROR(sp = e_string_new(0,sip->sys_file.FileName,E_PATH_SIZE));
    E_NULL_ERROR(sp = e_string_append_buffer(sp,".DIST")); 
  
    /* if system.DIST does not exists, create it */
    if(!e_stat_file(E_STRING_S(sp)))
    { E_NULL_ERROR(rename_file(sip->sys_file.FileName,E_STRING_S(sp),echo));

      mprintf("Creating the .DIST file %s\n",E_STRING_S(sp));
    }
    E_FREE(sp);
    sp = 0;
  }

  if(sip->config_file.none)
  { /* remove the system file if it exists */
    /* If the .DIST file was created, the system file has been moved. */
    if(!sip->sys_file.exists || !(sip->flags & E_SYSF_FLAG_r)) goto end;

    if (sip->flags & E_SYSF_FLAG_d)
    { E_NULL_ERROR(remove_dir(sip->sys_file.FileName,echo));
    }
    else
    { E_NULL_ERROR(remove_file(sip->sys_file.FileName,echo));
    }
  }
  else if (sip->flags & E_SYSF_FLAG_l)
  {            /* create or update the symbolic link */
    if(!sip->sys_file.exists)
    { mprintf("Creating the system file %s\n",sip->sys_file.FileName);
    }

    E_NULL_ERROR(remove_file(sip->sys_file.FileName,echo));

    E_NULL_ERROR(make_sym_link(sip->config_file.LinkPath,
                                    sip->sys_file.FileName,echo));
  }
  else if (sip->flags & E_SYSF_FLAG_d)
  {            /* create the directory recursively */
    if(!sip->sys_file.exists) 
    { mprintf("Creating the system directory %s\n",sip->sys_file.FileName);
      if(!mk_path_dir(sip->sys_file.FileName,echo)) goto error;
      sip->sys_file.exists = 1;
    }

    install_file_props(sip,1,1,1,echo);
  }
  else if (sip->flags & E_SYSF_FLAG_g)
  {           /* create the directory recursively */
    if(!sip->sys_file.exists)
    { mprintf("Creating the system directory (DIR LIST) %s\n",
               sip->sys_file.FileName);
      if(!mk_path_dir(sip->sys_file.FileName,echo)) goto error;
      sip->sys_file.exists = 1;
              /* we now reconfig the DIR LIST */

      E_NULL_ERROR(sip->dir_sys_asp = 
             dir_read_sys(sip->sys_file.FileName,sip->dir_config_asp));

      if (sip->dir_sys_asp && sip->dir_config_asp)
      { sip->diff_asp = dir_list_diff(sip->dir_sys_asp,sip->dir_config_asp);
      }
      else sip->diff_asp = 0;
    }

              /* install the dir list */
    if(!dir_install(sip,sip->diff_asp,sip->sys_file.FileName,echo,
                    INSTALL_MODE_FIRST_INSTALL))
    { mprintf("Unable to fully install the DIR LIST  %s\n",
              sip->sys_file.FileName);
    }
  }
  else
  { if(!sip->sys_file.exists)
    { mprintf("Creating the system file %s\n",sip->sys_file.FileName);
    }

    E_NULL_ERROR(copy_file(sip->config_file.FileName,
                                 sip->sys_file.FileName,echo));
  }

end:
  if(sp) E_FREE(sp);
  return(1);

error:
  if(sp) E_FREE(sp);
  return(0);
}

/******************************************************************/
/* Copy the system file into the config file.                     */
/* The system file must exist.                                    */
/* The config file must already exist.                            */
/* Nothing happens if the config file is up to date.              */
/* Only .HOST config files are Archived.                          */
/* name : config name                                             */
/* sip : pointer to sysfile_file info structure                   */
/******************************************************************/
int archive(char *name,sysfile_info *sip,int echo)
{ e_string *sp = 0;
  FILE *fpt;

  if(sip->config_file.exists)
  {         /* .DEFAULT file : use -makedef command */
    if(sip->config_file.config_file_type & E_CONFIG_DEFAULT)
    { mprintf("Use the \"-makedef\" command to update %s\n",
               sip->config_file.FileName);
    }
            /* .SHARED file : use -makeshared command */
    else if(sip->config_file.config_file_type & E_CONFIG_SHARED)
    { mprintf("Use the \"-makeshared\" command to update %s\n",
               sip->config_file.FileName);
    }
            /* .GROUP file : use -makedef command */
    else if(sip->config_file.config_file_type & E_CONFIG_GROUP)
    { mprintf("Use the \"-makegroup\" command to update %s\n",
               sip->config_file.FileName);
    }
    else    /* Archive the file */ 
    {       /* do not archive if there is no difference */
      if(sip->differ || sip->mode_differ || !sip->sys_file.exists)
      { if (!sip->sys_file.exists)
	{ /* No System file : set config to NONE */
          E_NULL_ERROR(make_sym_link("NONE",sip->config_file.FileName,echo));
	}
        else if(!test_file_type(sip->flags,&sip->sys_file,0,0))
        { return(1);
        }
        else if (sip->flags & E_SYSF_FLAG_l)
        { /* Remove the old config file in case it is a symbolic link */
          E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

          /* Write the link path into the config file */
          E_NULL_ERROR(fpt=e_open_file(sip->config_file.FileName,"w"));
  	  fprintf(fpt,"%s\n",sip->sys_file.LinkPath);
          fclose(fpt);
        }
        else if (sip->flags & E_SYSF_FLAG_d)
        { /* Remove the old config file in case it is a symbolic link */
          E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

          /* touch the config file */
          if(!mknod(sip->config_file.FileName,S_IFREG|0644,0))
	      sip->config_file.exists = 1;

          E_NULL_ERROR(archive_file_props(sip,1,1,1,echo));

        }
        else if (sip->flags & E_SYSF_FLAG_g)
        { /* Remove the old config file in case it is a symbolic link */
          E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

          /* Create the config file from the system data */
          if(sip->dir_sys_asp
             && dir_make_config_file(sip->dir_sys_asp,
                                  sip->config_file.FileName,echo))
	      sip->config_file.exists = 1;

        }
        else if(sip->differ)
        { /* Remove the old config file in case it is a symbolic link */
          E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

          E_NULL_ERROR(copy_file(sip->sys_file.FileName,
                                       sip->config_file.FileName,echo));
	}
        else
        { /* copy mode and ownership */
          E_NULL_ERROR(archive_file_props(sip,1,1,1,echo));
	}
      }
    }
  }
  else             /* File does not exist, use \"-new\" flag instead */
  { mprintf(
     "There is no Config file for %s.\nUse the -new command to update it.\n",
             name);
  }

  if(sp) E_FREE(sp);
  return(1);
error:
  if(sp) E_FREE(sp);
  return(0);
}

/******************************************************************/
/* Copy the system file as a new config file : GROUP/CONFIG.HOSTS */
/* The config file must not exist yet.                            */
/* name : config name                                             */
/* sip : pointer to sysfile_file info structure                   */
/******************************************************************/
int new(char *name,sysfile_info *sip,int echo)
{ FILE *fpt;

  if(!sip->config_file.exists)
  { if (!sip->sys_file.exists)
    { /* No System file : set config to NONE */
      E_NULL_ERROR(make_sym_link("NONE",sip->config_file.FileName,echo));
    }
    else if(!test_file_type(sip->flags,&sip->sys_file,0,0))
    { return(1);
    }
    else if (sip->flags & E_SYSF_FLAG_l)
    { /* Remove the old config file in case it is a symbolic link */
      E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

      /* Write the link path into the config file */
      E_NULL_ERROR(fpt=e_open_file(sip->config_file.FileName,"w"));
      fprintf(fpt,"%s\n",sip->sys_file.LinkPath);
      fclose(fpt);
    }
    else if (sip->flags & E_SYSF_FLAG_d)
    { /* Remove the old config file in case it is a symbolic link */
      E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

      /* touch the config file */
      if(!mknod(sip->config_file.FileName,S_IFREG|0644,0))
         sip->config_file.exists = 1;
      E_NULL_ERROR(archive_file_props(sip,1,1,1,echo));

    }
    else if (sip->flags & E_SYSF_FLAG_g)
    { /* Remove the old config file in case it is a symbolic link */
      E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

          /* Create the config file from the system data */
      if(sip->dir_sys_asp
         && dir_make_config_file(sip->dir_sys_asp,
                              sip->config_file.FileName,echo))
         sip->config_file.exists = 1;

    }
    else
    { /* Remove the old config file in case it is a symbolic link */
      E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

      E_NULL_ERROR(copy_file(sip->sys_file.FileName,
                             sip->config_file.FileName,echo));
    }
  }
  else             /* File exists, use Archive instead */
  { mprintf("Config file %s exists.\nUse the Archive command to update it.\n",
             sip->config_file.FileName);
  }

  return(1);

error:
  return(0);
  
}

/**************************************************************************/
/* Copy the system file as a specific config file:                        */
/* the DEFAULT config file : GROUP/CONFIG.DEFAULT                         */
/* the SHARED config file : GROUP/CONFIG.SHARED                           */
/* HOST_GROUP config file : GROUP/CONFIG.HOST_GROUP                       */
/* If the config file does not exist, it is created.                      */
/* name : config name                                                     */
/* sip : pointer to sysfile_file info structure                           */
/**************************************************************************/
int makedef(char *name,sysfile_info *sip,int echo)
{ FILE *fpt;

  if (!sip->sys_file.exists)
  { /* No System file : set config to NONE */
    E_NULL_ERROR(make_sym_link("NONE",sip->config_file.FileName,echo));
  }
  else if(!test_file_type(sip->flags,&sip->sys_file,0,0))
  { return(1);
  }
  else if (sip->flags & E_SYSF_FLAG_l)
  { /* Remove the old config file in case it is a symbolic link */
    E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

    /* Archive the symbolic link */
    /* Write the link path into the config file */
    E_NULL_ERROR(fpt=e_open_file(sip->config_file.FileName,"w"));
    fprintf(fpt,"%s\n",sip->sys_file.LinkPath);
    fclose(fpt);
  }
  else if (sip->flags & E_SYSF_FLAG_d)
  { /* Remove the old config file in case it is a symbolic link */
    E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

    /* touch the config file */
    if(!mknod(sip->config_file.FileName,S_IFREG|0644,0))
         sip->config_file.exists = 1;

    E_NULL_ERROR(archive_file_props(sip,1,1,1,echo));

  }
  else if (sip->flags & E_SYSF_FLAG_g)
  { /* Remove the old config file in case it is a symbolic link */
    E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

        /* Create the config file from the system data */
    if(sip->dir_sys_asp
       && dir_make_config_file(sip->dir_sys_asp,
                            sip->config_file.FileName,echo))
       sip->config_file.exists = 1;
  }
  else
  { /* Remove the old config file in case it is a symbolic link */
    E_NULL_ERROR(remove_file(sip->config_file.FileName,echo));

    E_NULL_ERROR(copy_file(sip->sys_file.FileName,
                             sip->config_file.FileName,echo));
  }

  return(1);

error:
  return(0);
}

/******************************************************************/
/* Display the differences between the config and the system file */
/* name : config name                                             */
/* sip : pointer to sysfile_file info structure                   */
/******************************************************************/
int diff(char *name,sysfile_info *sip,int echo)
{
  if(sip->config_file.none && !sip->sys_file.exists) return(1);

  if(!sip->config_file.exists)
  { mprintf("No config file for %s.\n",name);
    return(1);
  }

  if(!sip->sys_file.exists)
  { if (sip->flags & E_SYSF_FLAG_d)
       mprintf("System DIR %s does not exist.\n",sip->sys_file.FileName);
    else
       mprintf("System file %s does not exist.\n",sip->sys_file.FileName);
    return(1);
  }

  if(!test_file_type(sip->flags,&sip->sys_file,&sip->config_file,0))
  { return(1);
  }

  if (sip->flags & E_SYSF_FLAG_l)
  { if(strcmp(sip->sys_file.LinkPath,sip->config_file.LinkPath))
    { mprintf("System file symbolic link: %s\n",sip->sys_file.LinkPath);
      mprintf("Config file symbolic link: %s\n",sip->config_file.LinkPath);
    }
    return(1);
  }

  if (sip->flags & E_SYSF_FLAG_d)
  { /* if the directory exists, that's it! */
    return(1);
  }

  if (sip->flags & E_SYSF_FLAG_g)
  { dir_diff(sip->diff_asp,echo);
    return(1);
  }


               /* Execute the diff command   */
  E_NULL_ERROR(diff_files(sip->sys_file.FileName,
                          sip->config_file.FileName,name,echo));
  return(1);

error:
  return(0);
}

/******************************************************************/
/* Display the differences between the config files of two hosts  */
/* name : config name                                             */
/* fip : file_info pointer for the host being tested.             */
/* host1 : name of first host (argument of diff_host)             */
/* sip : pointer to sysfile_file info structure                   */
/* host2 : name of second host (current host)                     */
/******************************************************************/
int diff_host(char *name,file_info *fip,char *host1,sysfile_info *sip,
              char *host2,int echo)
{ 
  if(!fip->exists || !sip->config_file.exists)
  { if(!fip->exists && !sip->config_file.exists)
    { if(echo) mprintf("No config file %s for %s and %s.\n",name,host1,host2);
    }
    else
    { if(!fip->exists)
      { mprintf("No config file %s for host %s.\n",name,host1);
      }
      if(!sip->config_file.exists)
      { mprintf("No config file %s for host %s.\n",name,host2);
      }
    }
    return(1);
  }


  if(!strcmp(fip->FileName,sip->config_file.FileName))
  { /* same config file! */
    return(1);
  }

  if(fip->none && sip->config_file.none)
  { /* both set to NONE */
    return(1);
  }

  if (sip->flags & E_SYSF_FLAG_l)
  { if(fip->none)
    { mprintf("%\n",name);
      mprintf("%s : SYMBOLIC LINK: [NONE]\n",host1);
      mprintf("%s : SYMBOLIC LINK: %s\n",host2,sip->config_file.LinkPath);
    }
    
    else if(sip->config_file.none) 
    { mprintf("%\n",name);
      mprintf("%s : SYMBOLIC LINK: %s\n",host1,fip->LinkPath);
      mprintf("%s : SYMBOLIC LINK: [NONE]\n",host2);
    }   

    else if(strcmp(fip->LinkPath,sip->config_file.LinkPath))
    { mprintf("%\n",name);
      mprintf("%s : SYMBOLIC LINK: %s\n",host1,fip->LinkPath);
      mprintf("%s : SYMBOLIC LINK: %s\n",host2,sip->config_file.LinkPath);
    }
    return(1);
  }

  if (sip->flags & E_SYSF_FLAG_d)
  { if(fip->none)
    { mprintf("%s\n",name);
      mprintf("Config DIRECTORY set to NONE for %s but not for %s\n",
               host1,host2);
    }
    else if(sip->config_file.none)
    { mprintf("%s\n",name);
      mprintf("Config DIRECTORY set to NONE for %s but not for %s\n",
               host2,host1);
    }
    return(1);
  }

               /* Execute the diff command   */
  E_NULL_ERROR(diff_files(fip->FileName,sip->config_file.FileName,name,echo));

  return(1);

error:
  return(0);
}

/******************************************************************/
/* Check the differences between the config files of two hosts    */
/* name : config name                                             */
/* fip : file_info pointer for the host being tested.             */
/* host1 : name of first host (argument of check_host)            */
/* sip : pointer to sysfile_file info structure                   */
/* host2 : name of second host (current host)                     */
/******************************************************************/
int check_host(char *name,file_info *fip,char *host1,sysfile_info *sip,
               char *host2,int echo)
{ char host1_time[E_LINE_SIZE],host2_time[E_LINE_SIZE];
  int test;
 
  if(!fip->exists || !sip->config_file.exists)
  { if(!fip->exists && !sip->config_file.exists)
    { if(echo) mprintf("No config file %s for %s and %s.\n",name,host1,host2);
    }
    else
    { if(!fip->exists)
      { if(echo) mprintf("No config file %s for host %s.\n",name,host1);
      }
      if(!sip->config_file.exists)
      { if(echo) mprintf("No config file %s for host %s.\n",name,host2);
      }
    }
    return(1);
  }

  if(!strcmp(fip->FileName,sip->config_file.FileName))
  { /* same config file! */
    return(1);
  }

  if(fip->none && sip->config_file.none)
  { /* both set to NONE */
    return(1);
  }

  if (sip->flags & E_SYSF_FLAG_l)
  { if(fip->none)
    { mprintf("%\n",name);
      mprintf("%s : SYMBOLIC LINK: [NONE]\n",host1);
      mprintf("%s : SYMBOLIC LINK: %s\n",host2,sip->config_file.LinkPath);
    }
    
    else if(sip->config_file.none) 
    { mprintf("%\n",name);
      mprintf("%s : SYMBOLIC LINK: %s\n",host1,fip->LinkPath);
      mprintf("%s : SYMBOLIC LINK: [NONE]\n",host2);
    }   

    else if(strcmp(fip->LinkPath,sip->config_file.LinkPath))
    { mprintf("%\n",name);
      mprintf("%s : SYMBOLIC LINK: %s\n",host1,fip->LinkPath);
      mprintf("%s : SYMBOLIC LINK: %s\n",host2,sip->config_file.LinkPath);
    }
    return(1);
  }

  if (sip->flags & E_SYSF_FLAG_d)
  { if(fip->none)
    { mprintf("%s\n",name);
      mprintf("Config DIRECTORY set to NONE for %s but not for %s\n",
              host1,host2);
    }
    else if(sip->config_file.none)
    { mprintf("%s\n",name);
      mprintf("Config DIRECTORY set to NONE for %s but not for %s\n",
              host2,host1);
    }
    return(1);
  }


  if((test = files_are_equal(fip->FileName,sip->config_file.FileName))<0) 
  { mprintf("ERROR reading %s or %s.\n",fip->FileName,
            sip->config_file.FileName);
    return(0);
  }

  if(!test) /* file differ */
  { strftime(host1_time,E_LINE_SIZE,"%H:%M:%S %d/%m/%Y",
             localtime(&fip->time));
    strftime(host2_time,E_LINE_SIZE,"%H:%M:%S %d/%m/%Y",
             localtime(&sip->config_file.time));

    sysfile.CheckCounter++;
    mprintf("%s:\n",name);
    mprintf(" Config files Differ:\n",name);
    mprintf(" %s Config file (%s) : %s (%d)\n",host1,host1_time,fip->FileName,
                             fip->size);
    mprintf(" %s Config file (%s) : %s (%d)\n",host2,host2_time,
                             sip->config_file.FileName,sip->config_file.size);   
  }

  return(1);

}

/*****************************************************************/
/* Display the config file                                       */
/* name : config name                                            */
/* sip : pointer to sysfile_file info structure                  */
/*****************************************************************/
int cat_config(char *name,sysfile_info *sip,int echo)
{ FILE *fpt;
  char Buff[E_LONG_BUFF_SIZE];

  if(!sip->config_file.exists)
  { mprintf("No config file for %s.\n",name);
    return(1);
  }

  if(sip->config_file.none)
  { mprintf("%s is set to NONE\n");
    return(1);
  }

  if (sip->flags & E_SYSF_FLAG_l)
  { mprintf("%s :  Symbolic link: %s\n",name,sip->config_file.LinkPath);
    return(1);
  }

  if (sip->flags & E_SYSF_FLAG_d)
  { mprintf("%s :  Directory: %s\n",name,sip->config_file.FileName);
    return(1);
  }

  if(!(fpt = fopen(sip->config_file.FileName,"r")))
  { mprintf("Unable to open config file %s!",sip->config_file.FileName);
    return(1);
  }

  while(fgets(Buff,E_LONG_BUFF_SIZE,fpt))
  { mprintf("%s",Buff);
  }
  mprintf("\n");

  fclose(fpt);

  return(1);
}

/*****************************************************************/
/* Display the system file                                       */
/* name : config name                                            */
/* sip : pointer to sysfile_file info structure                  */
/*****************************************************************/
int cat_system(char *name,sysfile_info *sip,int echo)
{ FILE *fpt;
  char Buff[E_LONG_BUFF_SIZE];

  if(!sip->sys_file.exists)
  { mprintf("ERROR: System file does not exist.\n");
    return(1);
  }

  if(!test_file_type(sip->flags,&sip->sys_file,&sip->config_file,0))
  { return(1);
  }

  if (sip->flags & E_SYSF_FLAG_l)
  { mprintf("%s : Symbolic link: %s\n",name,sip->sys_file.LinkPath);
    return(1);
  }

  if (sip->flags & E_SYSF_FLAG_d)
  { mprintf("%s :  Directory: %s\n",name,sip->sys_file.FileName);
    return(1);
  }

  if (sip->flags & E_SYSF_FLAG_g)
  { dir_make_config_file(sip->dir_sys_asp,0,echo);
    return(1);
  }

  if(!(fpt = fopen(sip->sys_file.FileName,"r")))
  { mprintf("Unable to open system file %s!",sip->sys_file.FileName);
    return(1);
  }

  while(fgets(Buff,E_LONG_BUFF_SIZE,fpt))
  { mprintf("%s",Buff);
  }
  mprintf("\n");

  fclose(fpt);

  return(1);
}

/**************************************************************/
/* Copy the file mode, uid and gid from config to system file */
/* sip : pointer to sysfile_file info structure               */
/* do_uid : if non null : copy the uid.                       */
/* do_gid : if non null : copy the gid.                       */
/* do_mode : if non null : copy the file mode.                */
/* echo : echo flag                                           */
/* RETURN : 1 OK                                              */
/*        : 0 ERROR                                           */
/**************************************************************/
int install_file_props(sysfile_info *sip,int do_mode,int do_uid,int do_gid,
                       int echo)
{ char buff[10];

  if ((sip->flags & E_SYSF_FLAG_l) || (sip->flags & E_SYSF_FLAG_g))
  { /* No need to copy file ownership for symbolic links dir lists */
    return(1);
  }

  if(!sip->sys_file.exists)
  { mprintf("ERROR: System file %s does not exist.\n",sip->sys_file.FileName);
    if(!sip->config_file.exists) 
             mprintf("       Config file %s does not exist.\n",
                      sip->config_file.FileName);
    return(0);
  }
  if(!sip->config_file.exists)
  { mprintf("ERROR: Config %s file does not exist.\n",
            sip->config_file.FileName);
    return(0);
  }

  if(sip->config_file.none)
  { mprintf("ERROR: Config file %s set to NONE. 
       Can't use its attributes!\n",sip->config_file.FileName);
    return(0);
  }

  if(echo && do_uid) 
  { get_user_name(sip->config_file.uid,buff);
    mprintf("Changing uid of %s to %s\n",sip->sys_file.FileName,buff);
  }
  if(echo && do_gid) 
  { get_group_name(sip->config_file.gid,buff);
    mprintf("Changing gid of %s to %s\n",sip->sys_file.FileName,buff);
  }
  if(echo && do_mode) 
  { mprintf("Changing mode of %s to %o\n",sip->sys_file.FileName,
              sip->config_file.mode);
  }

  if(do_uid && do_gid)
  { if(chown(sip->sys_file.FileName,sip->config_file.uid,sip->config_file.gid))
    { mprintf("Unable to change the uid and gid for %s!",
              sip->sys_file.FileName);
      return(0);
    }
  }
  else 
  { if(do_uid) 
    { if (chown(sip->sys_file.FileName,sip->config_file.uid,-1))
      { mprintf("Unable to change the uid for %s!",
                sip->sys_file.FileName);
        return(0);
      }
    }

    if(do_gid) 
    { if (chown(sip->sys_file.FileName,-1,sip->config_file.gid))
      { mprintf("Unable to change the gid for %s!",
                sip->sys_file.FileName);
        return(0);
      }
    }
  }

  if(do_mode)
  { if(chmod(sip->sys_file.FileName,sip->config_file.mode))
    { mprintf("Unable to change the file mode for %s!",
              sip->sys_file.FileName);
      return(0);
    }
  }

  return(1);
}

/**************************************************************/
/* Copy the file mode, uid and gid from system to config file */
/* sip : pointer to sysfile_file info structure               */
/* do_uid : if non null : copy the uid.                       */
/* do_gid : if non null : copy the gid.                       */
/* do_mode : if non null : copy the file mode.                */
/* echo : echo flag                                           */
/* RETURN : 1 OK                                              */
/*        : 0 ERROR                                           */
/**************************************************************/
int archive_file_props(sysfile_info *sip,int do_mode,int do_uid,int do_gid, 
                       int echo)
{ char buff[10];

  if ((sip->flags & E_SYSF_FLAG_l)||(sip->flags & E_SYSF_FLAG_g))
  { /* No need to copy file ownership for symbolic links */
    return(1);
  }

  if(!sip->sys_file.exists)
  { mprintf("ERROR: System file %s does not exist.\n",
            sip->sys_file.FileName);
    if(!sip->config_file.exists) 
             mprintf("       Config file %s does not exist.\n",
                             sip->config_file.FileName);
    return(0);
  }
  if(!sip->config_file.exists)
  { mprintf("ERROR: Config file %s does not exist.\n",
            sip->config_file.FileName);
    return(0);
  }

  if(sip->config_file.none)
  { mprintf("ERROR: Config file %s set to NONE. 
       Can't change its attributes!\n",sip->config_file.FileName);
    return(0);
  }

  if(echo && do_uid) 
  { get_user_name(sip->sys_file.uid,buff);
    mprintf("Changing uid of %s to %s\n",sip->config_file.FileName,buff);
  }
  if(echo && do_gid) 
  { get_group_name(sip->sys_file.gid,buff);
    mprintf("Changing gid of %s to %s\n",sip->config_file.FileName,buff);
  }
  if(echo && do_mode) 
  { mprintf("Changing mode of %s to %o\n",sip->config_file.FileName,
              sip->sys_file.mode);
  }

  if(do_uid && do_gid)
  { if(chown(sip->config_file.FileName,sip->sys_file.uid,
       sip->sys_file.gid))
    { mprintf("Unable to change the uid and gid for %s!",
              sip->config_file.FileName);
      return(0);
    }
  }
  else 
  { if(do_uid) 
    { if(chown(sip->config_file.FileName,sip->sys_file.uid,-1))
      { mprintf("Unable to change the uid for %s!",
                sip->config_file.FileName);
        return(0);
      }
    }

    if(do_gid) 
    { if(chown(sip->config_file.FileName,-1,sip->sys_file.gid))
      { mprintf("Unable to change the gid for %s!",
                sip->config_file.FileName);
        return(0);
      }
    }
  }

  if(do_mode)
  { if(chmod(sip->config_file.FileName,sip->sys_file.mode))
    { mprintf("Unable to change the file mode for %s!",
              sip->config_file.FileName);
      return(0);
    }
  }

  return(1);
}

/************************************************************/
/* Check that each hosts in hosts belongs to the host_group */
/* Print "yes" if it does and "no" otherwise                */
/************************************************************/
void belongs_to_host_group(char *host_group,e_aset *hosts)
{ e_string *sp=0;
  e_aset *ashg=0;  
  int k,test = 1;

  E_NULL_ERROR(hosts);

  if(!(ashg = e_aset_get_value_by_key(sysfile.hosts_groups,
                                      e_name_new(host_group),0)))
  { mprintf("Error: unknown host group %s",host_group);
    goto error;
  }

  e_aset_next(hosts,E_NEXT_RESET,0);
  while(e_aset_next(hosts,E_NEXT_GET_NEXT,(void **)&sp) > 0)  
  { k = e_aset_get_key_index(ashg,e_name_new(E_STRING_S(sp)),0);
    if(k < 0) test = 0;
    E_FREE(sp);    
  }
  if(test == 1) mprintf("yes\n");
  else          mprintf("no\n");

  if(ashg)  E_FREE(ashg);

  return;

error:
  if(sp)    E_FREE(sp);
  if(ashg)  E_FREE(ashg);
}

/***********************************************************************/

/*************************************************/
/* Get the path name for the system file.        */
/* Get the last modification time and file size  */
/* Get the owner, group and access mode          */
/* Check that it is a file, not a directory.     */
/* If file is a symbolic link set fp->sym_link   */
/*  and copy path into fp->LinkPath              */
/* conf_name : config name                       */
/* conf_data : aset of data for this config file */
/* *fp : file_info to fill in                    */
/* flag : sys.config entry flags                 */
/* Return the result in fp.                      */
/* Return : 1 : file found                       */
/*          0 : system file does not exist.      */
/* Link conf_data : 0                            */
/*************************************************/
int get_system_file(char *conf_name,e_aset *conf_data,file_info *fp,int flag)
{ e_string *sp = 0;
  struct stat statp; 
  int i,j,c,stat_res;
  char *s;

  E_NULL_ERROR(fp);

                     /* Get system file name */
  E_NULL_ERROR(sp = e_aset_get_value_by_index(conf_data,CONFIG_SYSFILE_INDEX));

                     /* Copy the system file name in fp */
  /*
  strncpy(fp->FileName,E_STRING_S(sp),E_PATH_SIZE);
  fp->FileName[E_PATH_SIZE] = 0;
  */

  /* Copy filename and substitute $H with the host name */
  i = 0;
  s = E_STRING_S(sp);
  while((c = *s++) && (i < (E_PATH_SIZE-1)))
  { if(c == '$') /* variable : $H => hostname */
    { c = *s++;
      if(c != 'H')
      { mprintf("ERROR: Invalid variable ($%c) in system file name: %s!",
                 c,E_STRING_S(sp));
        goto error;
      }  
      j = 0;    /* copy hostname */
      while((fp->FileName[i] = HostName[j++]) && (i < (E_PATH_SIZE-1)))
      { i++;
      }
    }
    else fp->FileName[i++] = c;
  }
  fp->FileName[i] = 0;

               /* If the file exists, it must be a file */
  fp->exists = 0; 
  fp->type = 0;

  if(flag & E_SYSF_FLAG_l)
  { stat_res = lstat(fp->FileName,&statp); /* get link properties */
  }
  else
  { stat_res = stat(fp->FileName,&statp);
  }

  if(!stat_res)
  {  /* Check that the system file is of correct type */

    if(flag & (E_SYSF_FLAG_d|E_SYSF_FLAG_g))
    { if(!S_ISDIR(statp.st_mode))
      { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
                "System file %s IS NOT a DIRECTORY",fp->FileName);
        goto error;
      }
      fp->type |= E_SYSF_TYPE_DIR; 
    }
    else
    {  /* Check that the system file is a file */               
      if(!S_ISREG(statp.st_mode) && !S_ISLNK(statp.st_mode))
      { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
                "System file %s IS NOT a file",fp->FileName);
        goto error;
      }
    }
    fp->exists = 1; 

    if(flag & E_SYSF_FLAG_l)
    { if(S_ISLNK(statp.st_mode) && 
         ((i = readlink(fp->FileName,fp->LinkPath,E_PATH_SIZE)) >= 0))
      { fp->LinkPath[i] = 0;
        fp->type |= E_SYSF_TYPE_SYMLINK; 
      }
      else
      { fp->type &= ~E_SYSF_TYPE_SYMLINK; /* should be a symbolic link */
        fp->LinkPath[0] = 0;
      }
    }

                     /* Get the modification time and size */
    fp->time = statp.st_mtime;
    fp->size = statp.st_size;
    fp->uid = statp.st_uid;
    fp->gid = statp.st_gid;
    fp->mode = statp.st_mode;
  }

  if(sp) E_FREE(sp);

  return(1);

error :
  if(sp) E_FREE(sp);
  return(0);
}

/*****************************************************************/
/* Get the path name for the config file.                        */
/* Get the last modification time and file size                  */
/* Check that it is a file, not a directory.                     */
/* If config file is symlink to NONE, set fp->none               */
/* conf_name : config name                                       */
/* conf_data : aset of data for this config file                 */
/* *fp : file_info to fill in                                    */
/* host : the host name.                                         */
/* NoCheck : if non NULL, don't fail if the config file exists,  */
/*       just put the GROUP/CONFIG.HOST path in fp where         */
/*       host can be DEFAULT or a HOST_GROUP name                */
/*       Check if that file exists (for the new command)         */
/* flag : sys.config entry flags                                 */
/* Return the result in fp.                                      */
/* Return : 1 : file found                                       */
/*          0 : no config file for this host                     */
/* Link conf_data : 0                                            */
/*****************************************************************/
int get_config_file(char *conf_name,char *conf_dir,e_aset *conf_data,
                    file_info *fp,char *host,int NoCheck,int flag)
{ e_string *sp = 0,*config_name = 0,*group = 0,*conf_file_sp=0;
  e_name *np = 0;
  struct stat statp; 
  int found = 0,i;
  char Buff[E_PATH_SIZE],*config_file_prefix;

  if(!fp || !conf_data || !host) goto error;

  fp->exists = 0;
  fp->LinkPath[0] = 0;
  fp->FileName[0] = 0;

                     /* Get config name prefix */
  E_NULL_ERROR(conf_file_sp = 
        e_aset_get_value_by_index(conf_data,CONFIG_CONFFILE_INDEX));

  E_NULL_ERROR(config_file_prefix = E_STRING_S(conf_file_sp));

                     /* Get config file group */
  E_NULL_ERROR(group=e_aset_get_value_by_index(conf_data,CONFIG_GROUP_INDEX));

                /* sp : config file name :          */
                /* ConfigDir/Group/Config.EXTENSION */

  E_NULL_ERROR(sp = e_string_new(0,conf_dir,E_PATH_SIZE));
  E_NULL_ERROR(sp = e_string_append_buffer(sp,"/"));
  E_NULL_ERROR(sp = e_string_append_buffer(sp,E_STRING_S(group)));
  E_NULL_ERROR(sp = e_string_append_buffer(sp,"/"));
  E_NULL_ERROR(sp = e_string_append_buffer(sp,config_file_prefix));
  E_NULL_ERROR(sp = e_string_append_buffer(sp,"."));

  E_NULL_ERROR(config_name = e_string_append_buffer(E_USE(sp),host));
 
  if(NoCheck)
  {  strncpy(fp->FileName,E_STRING_S(config_name),E_PATH_SIZE);
     E_FREE(config_name); config_name = 0;

     fp->config_file_type = E_CONFIG_HOST; /* by default */
     if(!strcmp(host,"DEFAULT")) fp->config_file_type = E_CONFIG_DEFAULT;
     if(!strcmp(host,"SHARED"))  fp->config_file_type = E_CONFIG_SHARED;

     if(lstat(fp->FileName,&statp)) fp->exists = 0;
     else 
     { /* If config file is symlink to NONE : set none */
       fp->none = 0;
       if(S_ISLNK(statp.st_mode) && 
          ((i = readlink(fp->FileName,Buff,E_PATH_SIZE)) >= 0))
       { Buff[i] = 0;
         if(!strcmp(Buff,"NONE")) fp->none = 1;
       }

       /* symbolic link but not to NONE => stat the file pointed to */
       if(!fp->none && stat(fp->FileName,&statp))
       { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
   "sysf :  config file %s is a symbolic link to a file that does not exist!",
                 fp->FileName);
         goto error;
       }

       fp->exists = 1;
       fp->time = statp.st_mtime;
       fp->size = statp.st_size;
       fp->uid = statp.st_uid;
       fp->gid = statp.st_gid;
       fp->mode = statp.st_mode;

       if((flag & E_SYSF_FLAG_l) && !fp->none) /* get symlink path */
       { E_NULL_ERROR(read_first_line(fp->FileName,fp->LinkPath,E_PATH_SIZE));
         fp->type |= E_SYSF_TYPE_SYMLINK;
       }
       else if(flag & E_SYSF_FLAG_d)          /* get DIR path */
       { fp->type |= E_SYSF_TYPE_DIR;
       }
     }
     goto end;
  }


  fp->config_file_type = E_CONFIG_HOST; /* This is a .HOST config file */

  if(lstat(E_STRING_S(config_name),&statp))
  { E_FREE(config_name); config_name = 0;
 
                /* config.HOST does not exist, try each group in turn */   
    e_aset_next(sysfile.this_host_groups,E_NEXT_RESET,0);
  
    fp->config_file_type = E_CONFIG_GROUP; /* This is a .GROUP config file */

    while(!found && e_aset_next(sysfile.this_host_groups,E_NEXT_GET_NEXT,
                                (void **)&np)>0)
    { 
                                           /* does config.GROUP exist? */   
      E_NULL_ERROR(config_name = e_string_append_buffer(E_USE(sp),
                                                      E_NAME_NAME(np)));

      if(!lstat(E_STRING_S(config_name),&statp)) found = 1;
      else { E_FREE(config_name); config_name = 0; }
      if(np) E_FREE(np); 
      np = 0;
    }     

    if(!found)
    {                                    /* does config.DEFAULT exist? */ 
      E_NULL_ERROR(config_name = e_string_append_buffer(E_USE(sp),"DEFAULT"));
      if(lstat(E_STRING_S(config_name),&statp)) 
      { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
                "sysf on %s: no config file for %s",host,conf_name);
        goto error;
      }
      fp->config_file_type = E_CONFIG_DEFAULT; /* A .DEFAULT config file */
    }
  }

  strncpy(fp->FileName,E_STRING_S(config_name),E_PATH_SIZE);
  E_FREE(config_name); config_name = 0;

  /* Check that the system file is a file */    
  if(!S_ISREG(statp.st_mode) && !S_ISLNK(statp.st_mode) )
  { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
           "sysf :  config file %s is NOT A FILE",fp->FileName);
    goto error;
  }

  /* If config file is symlink to NONE : set none */
  fp->none = 0;
  if(S_ISLNK(statp.st_mode) && 
       ((i = readlink(fp->FileName,Buff,E_PATH_SIZE)) >= 0))
  { Buff[i] = 0;
    if(!strcmp(Buff,"NONE")) fp->none = 1;
  }

  /* symbolic link but not to NONE => stat the file pointed to */
  if(!fp->none && stat(fp->FileName,&statp))
  { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
   "sysf :  config file %s is a symbolic link to a file that does not exist!",
            fp->FileName);
    goto error;
  }

  if((flag & E_SYSF_FLAG_l) && !fp->none) /* get symlink path */
  { E_NULL_ERROR(read_first_line(fp->FileName,fp->LinkPath,E_PATH_SIZE));
    fp->type |= E_SYSF_TYPE_SYMLINK;
  }
  else if(flag & E_SYSF_FLAG_d)        /* get DIR path */
  { fp->type |= E_SYSF_TYPE_DIR;
  }

                     /* Get the modification time and size */
  fp->time = statp.st_mtime;
  fp->size = statp.st_size;
  fp->uid = statp.st_uid;
  fp->gid = statp.st_gid;
  fp->mode = statp.st_mode;
  fp->exists = 1;

end:
  if(sp) E_FREE(sp);
  if(group) E_FREE(group);

  return(1);

error :
  if(config_name) E_FREE(config_name);
  if(group) E_FREE(group);
  if(sp) E_FREE(sp);
  return(0);
}


/***********************************************/
/* Get the flags for the system file.          */
/* Return the result in *flag.                 */
/* Return : 1 : OK                             */
/*          0 : error                          */
/* Link conf_data : 0                          */
/***********************************************/
int get_system_file_flags(char *conf_name,e_aset *conf_data,int *flags)
{ e_string *flags_s = 0;
  int i,c;
  char *sp;

#ifdef DEBUG
  printf("get_system_file_flags(%s,%x,%x,%x)\n",conf_name,conf_data,flags);
#endif

  if(!conf_name || !conf_data || !flags) goto error;

  *flags = 0;
                     /* Get the flags if there is one */
  if(!(flags_s = e_aset_get_value_by_index(conf_data,CONFIG_FLAGS_INDEX))) 
     goto end;

  sp = E_STRING_S(flags_s);
  i = 0;
  while((c = sp[i]))
  { switch(c) 
    { case 'C' : *flags |= E_SYSF_FLAG_C;
               break;
      case 'I' : *flags |= E_SYSF_FLAG_I;
               break;
      case 'U' : *flags |= E_SYSF_FLAG_U;
               break;
      case 's' : *flags |= E_SYSF_FLAG_s;
               break;
      case 'h' : *flags |= E_SYSF_FLAG_h;
               break;
      case 'l' : *flags |= E_SYSF_FLAG_l;
               break;
      case 'd' : *flags |= E_SYSF_FLAG_d;
               break;
      case 'r' : *flags |= E_SYSF_FLAG_r;
               break;
      case 'g' : *flags |= E_SYSF_FLAG_g;
               break;
      case 'W' : *flags |= (E_SYSF_FLAG_W|E_SYSF_FLAG_C);
               break;
      default :
          mprintf("Config file %s : invalid flag %c\n",conf_name,c);
          goto error;
    }
    i++;
  }

end:
  if(flags_s) E_FREE(flags_s);

  return(1);

error :
  if(flags_s) E_FREE(flags_s);
  return(0);
}

/**************************************************************/
/* Get the path name for the auxiliary command if it exists.  */
/* Check that it is a file, not a directory.                  */
/* Return the result in fp.                                   */
/*         fp->aux_com.exists is 0 if command does not exist. */
/* Return : 1 : OK                                            */
/*          0 : error.                                        */
/* Link conf_data : 0                                         */
/**************************************************************/
int get_auxiliary_command(char *conf_name,char *conf_dir,e_aset *conf_data,
                       file_info *fp)
{ e_string *sp = 0,*sysname = 0;
  struct stat statp; 

  if(!conf_name || !conf_data || !fp) goto error;

                     /* Get system file name */
  E_NULL_ERROR(sysname = 
               e_aset_get_value_by_index(conf_data,CONFIG_GROUP_INDEX));

                     /* Copy the system command name in fp */
  E_NULL_ERROR(sp = e_string_new(0,conf_dir,E_PATH_SIZE));
  E_NULL_ERROR(sp = e_string_append_buffer(sp,"/BIN/"));
  E_NULL_ERROR(sp = e_string_append_buffer(sp,conf_name));

  strncpy(fp->FileName,E_STRING_S(sp),E_PATH_SIZE);
  fp->FileName[E_PATH_SIZE] = 0;

               /* If the file exists, it must be a file */
  if(!stat(fp->FileName,&statp))
  { 
                     /* Check that the auxiliary command is a file */
    if(!S_ISREG(statp.st_mode))
    { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
              "System command %s is a directory, not a file",fp->FileName);
      goto error;
    }

                     /* Check that the auxiliary command is executable */
    if(!((statp.st_mode & S_IXUSR) && (statp.st_mode & S_IXGRP)))
    { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
              "System command %s is not executable",fp->FileName);
      goto error;
    }
    fp->exists = 1; 
  }
  else fp->exists = 0; 

                     /* Get the modification time and size */
  fp->time = statp.st_mtime;
  fp->size = statp.st_size;

  if(sp) E_FREE(sp);
  if(sysname) E_FREE(sysname);

  return(1);

error :
  if(sp) E_FREE(sp);
  if(sysname) E_FREE(sysname);
  return(0);
}

/*************************************************/
/* Execute the system command file if it exists. */
/* SYS_COM ACTION COM SYSFILE_NAME CONFIG_FILE   */
/* The output is either displayed or emailed.    */
/* action : one of                               */
/*    -CreateSysFile                             */
/*    -UpdateSysFile                             */
/*    -PreInstall                                */
/*    -PostInstall                               */
/* command : sysf command (-install, ...)        */
/* fp : file info for this system file.          */
/* Return 1 : OK                                 */
/*        0 : error                              */
/*************************************************/
int execute_aux_com(char *action, char *command,sysfile_info *sip,int echo)
{ e_string *esp = 0;
  char *argv[7],*echo_s;

#ifdef DEBUG
printf("execute_aux_com(%s,%s,%x,%d)\n",action,command,sip,echo);
#endif

  if(!action || !command || !sip)
  { goto error;
  }

        /* No config file : no point to do anything */
  if(!sip->config_file.FileName[0])
  { return(1);
  }

  if(echo) echo_s = "on";
  else     echo_s = "off";

  if(!sip->aux_com.exists)
  { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
              "Command %s does not exist!",sip->aux_com.FileName);
    goto error;
  }
 
  argv[0] = sip->aux_com.FileName;
  argv[1] = action;
  argv[2] = command;
  argv[3] = sip->sys_file.FileName;
  argv[4] = sip->config_file.FileName;
  argv[5] = echo_s;
  argv[6] = 0;

  if((esp = e_exec_pipe_read(argv)))
  { mprintf("%s",E_STRING_S(esp));
    E_FREE(esp);
  }
  return(1);
  
error:
  if(esp) E_FREE(esp);
  return(0);
}

/********************************************************************/
/* Test if the sysfile and config file are identical and up to date */
/* sip : pointer to sysfile_file info structure                     */
/* Return : 1 sysfile and config file are identical                 */
/*          0 sysfile and config file differ                        */
/********************************************************************/
int test_sysfile(sysfile_info *sip)
{ 
  sip->differ = 0;
  sip->sys_file_older = 0;
  sip->mode_differ = 0;

  if(sip->flags & E_SYSF_FLAG_g)
  {       /* NONE for dir lists means ignore it */
    if(sip->config_file.none) return(1);

    if(sip->diff_asp)
      sip->differ = dir_config_sys_differ(sip->diff_asp);

    if(sip->sys_file.time < sip->config_file.time) sip->sys_file_older = 1;

    return(!sip->differ);
  }

  /* if sysf file must not exist */
  if(sip->config_file.none)
  { sip->differ = sip->sys_file.exists;
    return(!sip->differ);
  }

  if(sip->flags & E_SYSF_FLAG_l)
  { if(IS_SYMLINK(sip->sys_file.type))
    { sip->differ = 
	strcmp(sip->sys_file.LinkPath,sip->config_file.LinkPath) ? 1 : 0;
    }
    else sip->differ = 1;
    return(!sip->differ);
  } 
 
  if(sip->flags & E_SYSF_FLAG_d)
  { /* directories exists or they don't        */
    /* only the attributes need to be compared */
    if(!sip->sys_file.exists || ! sip->config_file.exists)
    { sip->differ = 1;
    }
  }
  else 
  { /* compare regular files */
    if (sip->sys_file.size != sip->config_file.size)
    { sip->differ = 1;
    }
    else
    { if(!files_are_equal(sip->sys_file.FileName,sip->config_file.FileName))
       sip->differ = 1;
    }
  }

  /* compare the attributes */
  if( ((sip->sys_file.mode && 02777) != (sip->config_file.mode && 02777)) ||
      (sip->sys_file.uid != sip->config_file.uid) ||
      (sip->sys_file.gid != sip->config_file.gid)
    )
  { sip->mode_differ = 1;
  }

  if(sip->sys_file.time < sip->config_file.time) sip->sys_file_older = 1;

  return(!sip->differ);
}

/*************************************************/
/* Test if the content of two files is identical */
/* Return : 1 : file are identical               */
/*        : 0 : files differ                     */
/*         -1 : error                            */
/*************************************************/
int files_are_equal(char *f1,char *f2)
{ int s1,s2,equal;
  FILE *fpt1,*fpt2;
  char Buff1[E_LONG_BUFF_SIZE],Buff2[E_LONG_BUFF_SIZE];

  if(!f1 || !f2) goto error;
  if(!(fpt1=fopen(f1,"r"))) goto error;
  if(!(fpt2=fopen(f2,"r"))) { fclose(fpt1); goto error;}
  
  equal = 1;
  do
  { s1 = fread(Buff1,1,E_LONG_BUFF_SIZE,fpt1);
    s2 = fread(Buff2,1,E_LONG_BUFF_SIZE,fpt2);
             /* files differ if the blocks are different */

    if(s1 != s2) equal = 0;
    else if(memcmp(Buff1,Buff2,s1)) equal = 0;
  }
  while(s1 && equal);

  fclose(fpt1);
  fclose(fpt2);
  return(equal);

error: 
  
  return(-1);
}

/*******************************************************/
/* Test that the file is of the correct type           */
/* If file does not exist, type is assumed correct!    */
/* flag : E_SYSF_FLAG_l : Symbolic link                */
/*        E_SYSF_FLAG_d : Directory                    */
/* sfip : system file file_info (if NULL -> ignore it) */
/* cfip : config file file_info (if NULL -> ignore it) */
/* mode : error message mode.                          */
/* Write an error message if there is a mismatch       */
/* Return : 1 : OK No problem, or file does not exist  */
/*          0 : Error, type mismatch                   */
/*******************************************************/
int test_file_type(int flag, file_info *sfip, file_info *cfip, int mode)
{ int r = 1;

  if(flag & E_SYSF_FLAG_l)
  { if(sfip && (sfip->exists) && !IS_SYMLINK(sfip->type))
    { mprintf("ERROR :  System file %s should be a SYMBOLIC LINK!\n",
               sfip->FileName);
      r = 0;
    }

    if(cfip && (cfip->exists) && !IS_SYMLINK(cfip->type))
    { mprintf("ERROR :  Config file %s should be a SYMBOLIC LINK!\n",
               cfip->FileName);
      r = 0;
    }
  }

  if(flag & E_SYSF_FLAG_d)
  { if(sfip && (sfip->exists) && !IS_DIR(sfip->type))
    { mprintf("ERROR :  System file %s should be a DIRECTORY!\n",
               sfip->FileName);
      r = 0;
    }

    if(cfip && (cfip->exists) && !IS_DIR(cfip->type))
    { mprintf("ERROR :  Config file %s should be a DIRECTORY!\n",
               cfip->FileName);
      r = 0;
    }
  }

  return(r);
}

/***********************************/
/* Read the first line from a file */
/* file : file name                */
/* buff : destination buffer       */
/* size : buffer size              */
/* Return : 1 : OK                 */
/*          0 error                */
/***********************************/
int read_first_line(char *file,char *buff,int size)
{ FILE *fpt = 0;
  int i,c;

  if(!(fpt = fopen(file,"r")))
  { e_error(E_ERROR_FILE ,E_ERROR_LEVEL_USER,
              "Can't open file %s",file);
    goto error;
  }

  i = 0;
  while(((c = fgetc(fpt))!= EOF) && (c != '\n') && (i < (size-1)))
  { buff[i++] = c;
  }
  buff[i] = 0;

  fclose(fpt);
  return(1);

error:
  if(fpt) fclose(fpt);
  return(0);
}

/********************************************************************/
/* Convert the uid into a user name.                                */
/* If the name can't be resolved, the uid is converted to a string. */
/* uid : the user identifier                                        */
/* buff : output buffer (at leat 9 chars)                           */
/* Return : Size of Buff.                                           */
/*          0 : error                                               */
/********************************************************************/
int get_user_name(int uid,char *buff)
{ struct passwd *pw;

  if((pw = getpwuid(uid)))
  { sprintf(buff,"%s",pw->pw_name);
  }
  else sprintf(buff,"%5d",uid);

  return(strlen(buff));
}

/********************************************************************/
/* Convert the gid into a group name.                               */
/* If the name can't be resolved, the uid is converted to a string. */
/* gid : the group identifier                                       */
/* buff : output buffer (at leat 9 chars)                           */
/* Return : Size of Buff.                                           */
/*          0 : error                                               */
/********************************************************************/
int get_group_name(int gid,char *buff)
{ struct group *gr;

  if((gr = getgrgid(gid)))
  { sprintf(buff,"%s",gr->gr_name);
  }
  else sprintf(buff,"%5d",gid);

  return(strlen(buff));
}

/************************** SMALL COMMANDS ************************/
/*****************************************************************/
/* mk_path_dir(path,echo) : creates a directory path recursively */
/* path : paths to create                                        */
/* echo : 1 : echo each action                                   */
/* Return : 1 : OK                                               */
/*          0 : error                                            */
/*****************************************************************/
int mk_path_dir(char *path,int echo)
{ char Com[E_LONG_BUFF_SIZE];
  e_string *sp = 0;
  int err;

  sprintf(Com,"%s -p %s",MKDIR_COM,path);
  if(echo) mprintf("%s\n",Com);

  if(!(sp = e_exec_sh_command(Com,&err))) goto error;
  mprintf("%s\n",E_STRING_S(sp));
  E_FREE(sp);

  return(1);

error:
  return(0);
}

/*****************************************************************/
/* remove_dir(path,echo) : removes the directory                 */
/*          % rmdir path                                         */
/* path : directory to remove                                    */
/* echo : 1 : echo each action                                   */
/* Return : 1 : OK                                               */
/*          0 : error                                            */
/*****************************************************************/
int remove_dir(char *path,int echo)
{ char Com[E_LONG_BUFF_SIZE];
  e_string *sp = 0;
  int err;

  sprintf(Com,"%s %s",RMDIR_COM,path);
  if(echo) mprintf("%s\n",Com);

  if(!(sp = e_exec_sh_command(Com,&err))) goto error;
  mprintf("%s\n",E_STRING_S(sp));
  E_FREE(sp);

  return(1);

error:
  return(0);
}

/*****************************************************************/
/* remove_file(path,echo) : removes the file                     */
/*          % rm -f path                                         */
/* path : file to remove                                         */
/* echo : 1 : echo each action                                   */
/* Return : 1 : OK                                               */
/*          0 : error                                            */
/*****************************************************************/
int remove_file(char *path,int echo)
{ char Com[E_LONG_BUFF_SIZE];
  e_string *sp = 0;
  int err;

  sprintf(Com,"%s -f %s",RM_COM,path);
  if(echo) mprintf("%s\n",Com);

  if(!(sp = e_exec_sh_command(Com,&err))) goto error;
  mprintf("%s\n",E_STRING_S(sp));
  E_FREE(sp);

  return(1);

error:
  return(0);
}

/*****************************************************************/
/* rename_file(path,new_path,echo) : rename the file             */
/*          % mv path new_path                                   */
/* path : file to rename                                         */
/* new_path : new file name                                      */
/* echo : 1 : echo each action                                   */
/* Return : 1 : OK                                               */
/*          0 : error                                            */
/*****************************************************************/
int rename_file(char *path,char *new_path,int echo)
{ char Com[E_LONG_BUFF_SIZE];
  e_string *sp = 0;
  int err;

  sprintf(Com,"%s %s %s",MV_COM,path,new_path);
  if(echo) mprintf("%s\n",Com);

  if(!(sp = e_exec_sh_command(Com,&err))) goto error;
  mprintf("%s\n",E_STRING_S(sp));
  E_FREE(sp);

  return(1);

error:
  return(0);
}

/*****************************************************************/
/* copy_file(path,new_path,echo) : copy the file with all the    */
/*                                  file attributes              */
/*          % cp -p path new_path                                */
/* path : file to rename                                         */
/* new_path : new file name                                      */
/* echo : 1 : echo each action                                   */
/* Return : 1 : OK                                               */
/*          0 : error                                            */
/*****************************************************************/
int copy_file(char *path,char *new_path,int echo)
{ char Com[E_LONG_BUFF_SIZE];
  e_string *sp = 0;
  int err;

  sprintf(Com,"%s -p %s %s",CP_COM,path,new_path);
  if(echo) mprintf("%s\n",Com);

  if(!(sp = e_exec_sh_command(Com,&err))) goto error;
  mprintf("%s\n",E_STRING_S(sp));
  E_FREE(sp);

  return(1);

error:
  return(0);
}

/*****************************************************************/
/* make_sym_link(path,link_path,echo) : create a symbolic link   */
/*          % ln -sf path link_path                              */
/* path : name for the file to link to                           */
/* link_path : name for the symbolic link                        */
/* echo : 1 : echo each action                                   */
/* Return : 1 : OK                                               */
/*          0 : error                                            */
/*****************************************************************/
int make_sym_link(char *path,char *link_path,int echo)
{ char Com[E_LONG_BUFF_SIZE];
  e_string *sp = 0;
  int err;

#ifdef SOLARIS
         /* ln -sf does not work on solaris => remove the file first */
  if(!remove_file(link_path,echo)) goto error;
#endif

  sprintf(Com,"%s -sf %s %s",LN_COM,path,link_path);
  if(echo) mprintf("%s\n",Com);

  if(!(sp = e_exec_sh_command(Com,&err))) goto error;
  mprintf("%s\n",E_STRING_S(sp));
  E_FREE(sp);

  return(1);

error:
  return(0);
}

/*****************************************************************/
/* diff_files(path1,path2,echo) : diff two files                 */
/*          % diff path1 path2                                   */
/* path : file to rename                                         */
/* link_path : new file name                                     */
/* name : sysf name (displayed before the diff output)           */
/* echo : 1 : echo each action                                   */
/* Return : 1 : OK                                               */
/*          0 : error                                            */
/*****************************************************************/
int diff_files(char *path1,char *path2,char *name,int echo)
{ char Com[E_LONG_BUFF_SIZE];
  e_string *sp = 0;
  int err;

  sprintf(Com,"diff %s %s",path1,path2);

  if(echo) mprintf("%s\n",Com);

               /* Execute the diff command */
  if(!(sp = e_exec_sh_command(Com,&err))) goto error;

  if(E_STRING_S(sp)[0])
  { mprintf("%s:\n%s",name,E_STRING_S(sp));
  }

  E_FREE(sp);

  return(1);

error:
  return(0);
}


