#include <sysf.h>
#include <sysf_dir.h>
#include <sysf_init.h>
#include <sysf_com.h>
#include <md5/md5.h>

#define NO_ENTRY_KEY 8
char *Dir_Entry_Keys[NO_ENTRY_KEY+1] = {
 "Name",
 "Owner",
 "Group",
 "Modes",
 "Type",
 "Size",
 "CheckSum",
 "Link",
 0,
};
e_name *Dir_Entry_Names[NO_ENTRY_KEY+1];

extern sysf_cf sysf_cf_vars;

/********************************/
/* Init the Dir Entry Name list */
/********************************/
void init_dir_entry_names()
{ int i;

  for(i = 0; i < NO_ENTRY_KEY; i++)
  { Dir_Entry_Names[i] = e_name_new(Dir_Entry_Keys[i]);
  }

}

/***********************************************************************/
/* Read a directory list into an aset                                  */
/*  Key : filename.                                                    */
/*  Value : { Name; Owner; Group; Modes; Type; Size; CheckSum; Link; } */
/* dir_path : path for the DIR LIST directory                          */
/* asp_config : config aset                                            */
/* Return : the directory aset.                                        */
/*          0 : error                                                  */
/* Link : Return : 1                                                   */
/***********************************************************************/
e_aset *dir_read_sys(char *dir_path,e_aset *asp_config)
{ DIR *dpt;
  struct dirent *dent;
  e_aset *as_dir_list=0,*as_file,*val;
  e_string *stp=0;
  e_name *np=0;
  char *type=0;

  if(!dir_path || !dir_path[0]) return(0);

  if(!S_ISDIR(e_stat_file(dir_path)) || !(dpt = opendir(dir_path)))
  { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
              "Can't open directory %s",dir_path);
    goto error;
  }
  E_NULL_ERROR(stp = e_string_new(0,0,E_BUFF_SIZE));
  E_NULL_ERROR(as_dir_list = e_aset_new(0));

  while((dent = readdir(dpt)) != NULL)
  { /* ignore the . and .. entries */
    if(strcmp(dent->d_name,".") && strcmp(dent->d_name,".."))
    { e_string_reset(stp,dir_path);
      E_NULL_ERROR(stp = e_string_append_buffer(stp,"/"));
      E_NULL_ERROR(stp = e_string_append_buffer(stp,dent->d_name));
      E_NULL_ERROR(np=e_name_new(dent->d_name));

      type =0;
      if(asp_config && (val = e_aset_get_value_by_key(asp_config,E_USE(np),0)))
      { type = e_aset_get_string_pointer_by_key(val,DE_NAME_TYPE);
        E_FREE(val);
      }

      /* append the file properties to as_dir_list */
      if((as_file = get_file_pars(E_STRING_S(stp),type)))
      { E_NULL_ERROR(as_dir_list = e_aset_append_item(as_dir_list,np,as_file));
        np=0;
      }
    }
  }
  closedir(dpt);
  E_SAFE_FREE(stp);

  as_dir_list = e_aset_sort_by_key(as_dir_list,E_MODE_INCREASING);

/*
  e_tools_print_aset("as_dir_list = ",as_dir_list,stdout);
*/

  return(as_dir_list);

error:
  E_SAFE_FREE(np);
  E_SAFE_FREE(stp);
  E_SAFE_FREE(as_dir_list);

  return(0);

}

/**************************************************************/
/* Read the property list of a file and return into an aset:  */
/*  { Name=STR; Owner=UINT; Group=UINT; Modes=INT; Type=STR;  */
/*    Size=INT; CheckSum=STR; Link=STR;  }                    */
/* path : path name for the file to stat                      */
/* conf_type : config type (preserve all but the lasts char)  */
/* Return : the directory aset.                               */
/*          0 : error                                         */
/* Link : Return : 1                                          */
/**************************************************************/
e_aset *get_file_pars(char *path,char *conf_type)
{ e_aset *asp=0;
  e_number *nup=0;
  e_string *strp=0;
  char *s;
  struct stat stp;
  char Buff[E_BUFF_SIZE],type[E_WORD_SIZE],checksum[E_BUFF_SIZE];
  int n;
  int size;

  s = path+e_tools_str_get_suffix(path,'/');

  E_NULL_ERROR(asp = e_aset_new(0));

  /* Name = */
  E_NULL_ERROR(asp =
            e_aset_append_item(asp,E_USE(DE_NAME_NAME),e_string_new(0,s,0))); 

  if(!lstat(path,&stp))
  { 
    /* Owner = */
    E_NULL_ERROR(nup = e_number_new(0,E_NUM_TYPE_UINT,stp.st_uid));
    E_NULL_ERROR(asp = e_aset_append_item(asp,E_USE(DE_NAME_OWNER),nup)); 
    nup = 0;

    /* Group = */
    E_NULL_ERROR(nup = e_number_new(0,E_NUM_TYPE_UINT,stp.st_gid));
    E_NULL_ERROR(asp = e_aset_append_item(asp,E_USE(DE_NAME_GROUP),nup)); 
    nup = 0;

    /* Modes = */
    E_NULL_ERROR(nup = 
           e_number_new(0,E_NUM_TYPE_UINT,stp.st_mode&FILE_MODE_MASK));
    E_NULL_ERROR(asp = e_aset_append_item(asp,E_USE(DE_NAME_MODES),nup)); 
    nup = 0;

    /* Type = */
    if(S_ISREG(stp.st_mode)) s = "f"; 
    else if(S_ISDIR(stp.st_mode)) s = "d"; 
    else if(S_ISCHR(stp.st_mode)) s = "c"; 
    else if(S_ISBLK(stp.st_mode)) s = "b"; 
    else if(S_ISFIFO(stp.st_mode)) s = "F"; 
    else if(S_ISLNK(stp.st_mode)) s = "l"; 
    else s="?";

    if(conf_type)  /* use conf_type except for the first char */
    { strncpy(type,conf_type,E_WORD_SIZE-1);
      type[0] = s[0];
      type[E_WORD_SIZE-1] = 0;
    }
    else strncpy(type,s,E_WORD_SIZE-2);

    E_NULL_ERROR(asp = 
          e_aset_append_item(asp,E_USE(DE_NAME_TYPE),e_string_new(0,type,0))); 

    /* Size = */
    size = 0;
    strcpy(checksum,"0");

    if(!strchr(type,'0') && !strchr(type,'I'))
    { /* get size and checksum unless type is 0 or I */
      if(S_ISREG(stp.st_mode))
      { size = stp.st_size;
               /* CheckSum (not if flag S is set) */
        if(!strchr(type,'S'))
        { sysf_check_sum(path,checksum,0);
        }
      }
    }
    E_NULL_ERROR(nup = e_number_new(0,E_NUM_TYPE_INT,size));
    E_NULL_ERROR(asp = e_aset_append_item(asp,E_USE(DE_NAME_SIZE),nup)); 
    nup = 0;

    E_NULL_ERROR(strp = e_string_new(0,checksum,0));
    E_NULL_ERROR(asp = e_aset_append_item(asp,E_USE(DE_NAME_CKSUM),strp));
    strp = 0;

    /* Link = */
    if(S_ISLNK(stp.st_mode))
    { if((n = readlink(path,Buff, E_BUFF_SIZE)) >= 0)
      { Buff[n] = 0;
        E_NULL_ERROR(asp = 
         e_aset_append_item(asp,E_USE(DE_NAME_LINK),e_string_new(0,Buff,0))); 
      }
    }
  }

  return(asp);

error:
  E_SAFE_FREE(asp);
  E_SAFE_FREE(strp);

  return(0);
}

/*******************************************************/
/* Create the directory config file file_name from the */
/* Directory list aset.                                */
/* Return : 1 : OK                                     */
/*          0 : error                                  */
/* Link : asp : 0                                      */
/*******************************************************/
int dir_make_config_file(e_aset *asp,char *file_name,int echo)
{ FILE *fpt;
  e_aset_item *aip;

  if(!file_name) fpt = 0; /* stdout */
  else 
  { E_NULL_ERROR(fpt=e_open_file(file_name,"w"));
  }

  aip = E_ASET_FIRST(asp);

  while(aip)
  { print_file_props(E_ASET_ITEM_VAL(aip),fpt);
    aip = aip->next;
  }

  if(fpt) fclose(fpt); 


  return(1);

error:
  return(0);

}

/*******************************************/
/* Display the file properties on one line */
/* Link : asp : 0                          */
/*******************************************/
void print_file_props(e_aset *asp,FILE *fpt)
{ e_aset_item *aip;
  e_string *fsp;
  e_number *nump;

  if(asp && E_IS_ASET(asp))
  { aip = E_ASET_FIRST(asp);
      
    /* Name */
    if(aip && (fsp = E_ASET_ITEM_VAL(aip)))
    { if(fpt) fprintf(fpt,"%s =",E_STRING_S(fsp));
      else mprintf("%s =",E_STRING_S(fsp));
    }

    while((aip = aip->next)) 
    {if(E_ASET_ITEM_NAME(aip) == DE_NAME_MODES)
      { /* this is an octal number */
        if((nump = E_ASET_ITEM_VAL(aip)))
        { if(fpt) fprintf(fpt," %o",(int) E_NUMBER_VAL_UINT(nump));
          else    mprintf(" %o",E_NUMBER_VAL_UINT(nump));
          E_FREE(nump);
        }
      }
      else if((fsp = E_TO_STRING(E_ASET_ITEM_VAL(aip),0,E_TO_STR_QUOTE)))
      { if(fpt) fprintf(fpt," %s",E_STRING_S(fsp));
        else    mprintf(" %s",E_STRING_S(fsp));
        E_FREE(fsp);
      }
    }
    if(fpt) fprintf(fpt,";\n");
    else    mprintf(";\n"); 
  }
}

/*****************************************************************************/
/* Read a directory config file and return it as a directory aset            */
/* Dir file format : sequence of                                             */
/*  NAME =  UID  GID  MODE  TYPE  SIZE  CheckSum [LINK];                     */
/* Output : aset of                                                          */
/* { name = { Name; Owner; Group; Modes; Type; Size; Link; CheckSum; }}      */
/* file_path : path for the file to read.                                    */
/* Return : e_aset :                                                         */
/*          0 : error                                                        */
/*****************************************************************************/
e_aset *dir_read_config(char *file_path)
{ e_tokeniser *tp = 0;
  e_aset *dir_asp = 0,*item_list = 0;
  e_name *item_name = 0;
  void *obj;
  int state,par_no=0,ival;
  unsigned int uval;
  e_flag mode = E_TOKENISER_MODE_HASH_COMMENT|E_TOKENISER_MODE_READ_STRING;

  if(!(tp = e_tokeniser_new(0,e_charstream_new(0,
                                 E_CHARSTREAM_TYPE_FILE,
                                 e_file_new(0,0,file_path,"r"),
                                 E_MODE_READ),
                            mode,E_SYSF_DIR_DEF_WORD,E_SYSF_DIR_DEF_OP,
                            E_SYSF_DIR_DEF_PUNCT,E_SYSF_DIR_DEF_SPACES)))
  { goto error;
  }

  E_NULL_ERROR(dir_asp = e_aset_new(0)); 

  state = STATE_DIR_NEW;
  while(e_tokeniser_next(tp) > 0)
  { switch(tp->type)
    { 
      case E_TOKENISER_STRING_TOKEN :
      case E_TOKENISER_WORD_TOKEN :
            if(state == STATE_DIR_NEW)         /* new entry */
            { E_NULL_ERROR((item_name = e_name_new(E_TOKENISER_BUFF(tp))));

              state = STATE_DIR_EQUAL;
              E_NULL_ERROR(item_list = e_aset_new(0)); 
              par_no = 0;

              E_NULL_ERROR((item_list = e_aset_append_item(item_list,
                               E_USE(Dir_Entry_Names[par_no]),
                               e_string_new(0,E_TOKENISER_BUFF(tp),0))));
              par_no++;
	    }          
            else if( state == STATE_DIR_EQUAL) /* Error  */
            { goto error1;
	    }          
            else                         /* add 1 the entry item */
	    { switch(par_no)
              { case DE_INDEX_OWNER:
                case DE_INDEX_GROUP:
                  sscanf(E_TOKENISER_BUFF(tp),"%u",&uval);
                  E_NULL_ERROR(obj = e_number_new(0,E_NUM_TYPE_UINT,0));
                  E_NUMBER_VAL_UINT(obj) = uval;
                  break;

                case DE_INDEX_MODES:
                  sscanf(E_TOKENISER_BUFF(tp),"%o",&uval);
                  E_NULL_ERROR(obj = e_number_new(0,E_NUM_TYPE_UINT,0));
                  E_NUMBER_VAL_UINT(obj) = uval;
                  break;

                case DE_INDEX_SIZE:
                  sscanf(E_TOKENISER_BUFF(tp),"%d",&ival);
                  E_NULL_ERROR(obj = e_number_new(0,E_NUM_TYPE_INT,ival));
                  break;

                default :
                  E_NULL_ERROR(obj = e_string_new(0,E_TOKENISER_BUFF(tp),0));
                  break;
              }

              E_NULL_ERROR((item_list = e_aset_append_item(item_list,
                               E_USE(Dir_Entry_Names[par_no]),obj
                               )));
              par_no++;
	    }        
            break;

      case E_TOKENISER_OP_TOKEN :       /* read the = op */
           if((E_TOKENISER_BUFF(tp)[0] != '=') || (state != STATE_DIR_EQUAL))
               goto error1;

            state = STATE_DIR_ITEM;
            break;

      case E_TOKENISER_PUNCT_TOKEN :       /* end of host list */
            if ((E_TOKENISER_BUFF(tp)[0] != ';') || (state != STATE_DIR_ITEM))
               goto error1;
            if(E_ASET_CARDINAL(item_list) < 6)
	    { e_error_syntax(
       "sysf: Syntax error reading %s file line %d:\n      To few parameter for item %s\n",
                  file_path,E_TOKENISER_LINE_NO(tp),E_NAME_NAME(item_name));
              E_FREE(item_name);
              goto error;
	    }            

	                          /* Append item list to the entry list */
            dir_asp = e_aset_append_item(dir_asp,item_name,item_list);
            item_list = 0; item_name = 0;

            state = STATE_DIR_NEW;

            break;
    }
  }

  if(tp) E_FREE(tp);

  return(dir_asp);

error1:
   e_error_syntax(
    "sysf: Syntax error reading %s file line %d: unexpected char %c\n", 
                 file_path,E_TOKENISER_LINE_NO(tp),E_TOKENISER_BUFF(tp)[0]);
error:
  if(tp)        E_FREE(tp);
  if(item_list) E_FREE(item_list);
  if(item_name) E_FREE(item_name);
  return(0);
}

/*****************************************************************/
/* Create and aset listing the differences between two dir lists */
/* The result is an aset with three named subset :               */
/*  Missing : item in asp2 but not in asp1                       */
/*  Extra : items in asp1 but not in asp2                        */
/*  Differ : list of item which have non matching properties     */
/* Return : aset = {missing={}; extra={}; differ={{}{}}}         */
/*          0 : error                                            */
/* Link : asp1 : 0                                               */
/* Link : asp2 : 0                                               */
/*****************************************************************/
e_aset *dir_list_diff(e_aset *asp1,e_aset *asp2)
{ e_aset_item *aip1;
  e_aset *diffs=0,*extra=0,*missing=0,*val1,*val2,*fasp1=0,*fasp2=0,*asp;
  e_name *np;
  char *type;

  if(!asp1 || !asp2) goto error;

  E_NULL_ERROR(fasp1 = E_CLONE(E_USE(asp1),0));
  E_NULL_ERROR(fasp2 = E_CLONE(E_USE(asp2),0));

  E_NULL_ERROR(missing = e_aset_new(0));
  E_NULL_ERROR(extra = e_aset_new(0));
  E_NULL_ERROR(diffs = e_aset_new(0));

  aip1 = E_ASET_FIRST(fasp1);

  while(aip1)
  { if((np = E_ASET_ITEM_NAME(aip1)))
    { val1 = E_ASET_ITEM_VAL(aip1);
      if ((val2 = e_aset_get_value_by_key(fasp2,E_USE(np),0)))
      { if(test_file_properties(val1,val2))
        { /* files differ : add to the diff list */
          E_NULL_ERROR(asp = make_aset(E_USE(val1),E_USE(val2),0));
          E_NULL_ERROR(diffs = e_aset_append_item(diffs,E_USE(np),asp));
        }

        E_SAFE_FREE(val2);
        
        E_NULL_ERROR(e_aset_remove_item_by_key(fasp1,np,1));
        E_NULL_ERROR(e_aset_remove_item_by_key(fasp2,np,1));
      }
      else /* add entry to extra list */
      { if(val1)
        { E_USE(val1);
          E_NULL_ERROR(extra = e_aset_append_item(extra,E_USE(np),val1));
        }
      }
    }
    aip1 = aip1->next;
  }
  E_SAFE_FREE(fasp1);

  /* loop through the missing list and remove item of type I */
  aip1 = E_ASET_FIRST(fasp2);
  while(aip1)
  { if((val1 = E_ASET_ITEM_VAL(aip1)) && E_IS_ASET(val1) && 
       (np = E_ASET_ITEM_NAME(aip1)))
    { type = e_aset_get_string_pointer_by_key(val1,DE_NAME_TYPE);
      if(!strchr(type,'I'))
       { E_USE(val1);
         E_NULL_ERROR(missing = e_aset_append_item(missing,E_USE(np),val1));
       }
    }
    aip1 = aip1->next;
  }
  E_SAFE_FREE(fasp2);

  return(make_named_aset(e_name_new("missing"),missing,
                         e_name_new("extra"),extra,
                         e_name_new("differ"),diffs,
                         0
                        ));


error:
  E_SAFE_FREE(fasp1);
  E_SAFE_FREE(fasp2);
  E_SAFE_FREE(missing);
  E_SAFE_FREE(diffs);
  E_SAFE_FREE(extra);
  return(0);
}

/*************************************************************/
/* Compare two file property lists                           */
/* owner, group and mode are not compared for symbolic links */
/* link is only tested for symbolic links                    */
/* size and checksum are only compared for files             */
/* Return : 0 : files are identical                          */
/*          1 : files differ                                 */
/* Link : asp1 0                                             */
/*        asp2 0                                             */
/*************************************************************/
int test_file_properties(e_aset *asp1,e_aset *asp2)
{ file_pars fp1,fp2;

  E_NULL_ERROR(fill_file_pars(asp1,&fp1));
  E_NULL_ERROR(fill_file_pars(asp2,&fp2));

              /* type I : ignore differences */
  if(strchr(fp2.type,'I')) return(0);

  if(strcmp(fp1.name,fp2.name)) return(1);
  if(fp1.type[0] != fp2.type[0]) return(1);

  if(fp1.type[0] != 'l')
  { if(fp1.modes != fp2.modes)    return(1);
    if(fp1.owner != fp2.owner)    return(1);
    if(fp1.group != fp2.group)    return(1);
  }
  else if(strcmp(fp1.link,fp2.link)) return(1);

  if((fp1.type[0] == 'f') && !strchr(fp1.type,'0'))
  { if(fp1.size != fp2.size)    return(1);
    if (!strchr(fp1.type,'S') && strcmp(fp1.checksum,fp2.checksum)) return(1);
  }
 
  return(0);

error:
  return(1);
}

/******************************************************/
/* Fill in the file_pars structure with the aset data */
/* Return : 1 : OK                                    */
/*          0 : ERROR                                 */
/* Link : asp : 0                                     */
/******************************************************/
int fill_file_pars(e_aset *asp,file_pars *fp)
{ e_string *stp=0;
  e_number *np=0;
  
  if(!asp || !fp) goto error;

  /* NAME */
  if((stp = e_aset_get_value_by_key(asp,E_USE(DE_NAME_NAME),0))
     && E_IS_STRING(stp))
  { strncpy(fp->name,E_STRING_S(stp),E_BUFF_SIZE-1);
    E_FREE(stp);
  }
  else fp->name[0] = 0;

  /* OWNER */
  if((np = e_aset_get_value_by_key(asp,E_USE(DE_NAME_OWNER),0))
     && E_IS_NUMBER(np))
  { fp->owner = E_NUMBER_VAL_UINT(np);
    E_FREE(np);
  }
  else fp->owner = 0;

  /* GROUP */
  if((np = e_aset_get_value_by_key(asp,E_USE(DE_NAME_GROUP),0))
     && E_IS_NUMBER(np))
  { fp->group = E_NUMBER_VAL_UINT(np);
    E_FREE(np);
  }
  else fp->group = 0;

  /* MODES */
  if((np = e_aset_get_value_by_key(asp,E_USE(DE_NAME_MODES),0))
     && E_IS_NUMBER(np))
  { fp->modes = E_NUMBER_VAL_UINT(np);
    E_FREE(np);
  }
  else fp->modes = 0;

  /* TYPE */
  if((stp = e_aset_get_value_by_key(asp,E_USE(DE_NAME_TYPE),0))
     && E_IS_STRING(stp))
  { strncpy(fp->type,E_STRING_S(stp),E_WORD_SIZE);
    E_FREE(stp);
  }
  else fp->type[0] = 0;

  fp->size = 0;
  fp->checksum[0] = 0;
  if(!strchr(fp->type,'0')) /* ignore size and checksum for type '0' */
  { /* SIZE */
    if((np = e_aset_get_value_by_key(asp,E_USE(DE_NAME_SIZE),0))
       && E_IS_NUMBER(np))
    { fp->size = E_NUMBER_VAL_INT(np);
      E_FREE(np);
    }

    /* CHECKSUM */
    if((!strchr(fp->type,'S')) && 
       (stp = e_aset_get_value_by_key(asp,E_USE(DE_NAME_CKSUM),0))
       && E_IS_STRING(stp))
    { strncpy(fp->checksum,E_STRING_S(stp),E_BUFF_SIZE);
      E_FREE(stp);
    }
  }

  /* LINK */
  if((stp = e_aset_get_value_by_key(asp,E_USE(DE_NAME_LINK),0))
     && E_IS_STRING(stp))
  { strncpy(fp->link,E_STRING_S(stp),E_BUFF_SIZE-1);
    E_FREE(stp);
  }
  else fp->link[0] = 0;

  return(1);

error:
  E_SAFE_FREE(stp);
  E_SAFE_FREE(np);
  return(0); 

}

/*********************************************************************/
/*********************************************************/
/* Test if the system and the config files are identical */
/* asp : { missing={};extra={}; differ{} }               */
/* Files differ if one of th 3 subset is not empty       */
/* return : 1 : sys and config differ                    */
/*          0 : sys and config match                     */
/* Link : asp : 0                                        */
/*********************************************************/
int dir_config_sys_differ(e_aset *asp)
{ e_aset_item *aip;
  e_aset *asp1;

  aip = E_ASET_FIRST(asp);
  while(aip)
  { asp1 = E_ASET_ITEM_VAL(aip);
    if(asp1 && (E_ASET_CARDINAL(asp1) > 0)) return(1); 
    aip = E_ASET_ITEM_NEXT(aip);
  }

  return(0);
}

/*********************************************************/
/* Display briefly the differences between sys           */
/* and config files by listing the keys of each set      */
/* asp : { missing={};extra={}; differ{{}{}} }           */
/* return : 1 : OK                                       */
/*          0 : error                                    */
/* Link : asp : 0                                        */
/*********************************************************/
int dir_check(e_aset *asp,int echo)
{ e_aset_item *aip,*aip1;
  e_aset *asp1;
  e_name *np;
  char *s=0;

  aip = E_ASET_FIRST(asp);
  while(aip)
  { if((np = E_ASET_ITEM_NAME(aip))) s = E_NAME_NAME(np);
    else s = 0;

    if((asp1 = E_ASET_ITEM_VAL(aip)) &&(E_ASET_CARDINAL(asp1) > 0 ))
    { mprintf("  %s :",s);
      aip1 = E_ASET_FIRST(asp1);
      while(aip1)
      { if((np = E_ASET_ITEM_NAME(aip1))) mprintf(" %s",E_NAME_NAME(np)); 
        aip1 = E_ASET_ITEM_NEXT(aip1);
      }
      mprintf("\n");
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }

  return(1);
}

/*********************************************************/
/* Display the differences between sys                   */
/* and config files by listing the properties of         */
/* each differing file                                   */
/* asp : { missing={};extra={}; differ{{}{}} }           */
/* return : 1 : OK                                       */
/*          0 : error                                    */
/* Link : asp : 0                                        */
/*********************************************************/
int dir_diff(e_aset *asp,int echo)
{ e_aset_item *aip,*aip1,*aip2;
  e_aset *asp1,*asp2;
  e_name *np;
  char *s=0;

  aip = E_ASET_FIRST(asp);
  while(aip)
  { if((np = E_ASET_ITEM_NAME(aip))) s = E_NAME_NAME(np);
    else s = 0;


    if((asp1 = E_ASET_ITEM_VAL(aip)) && (E_ASET_CARDINAL(asp1) > 0 ))
    { mprintf("  %s :\n",s);

      aip1 = E_ASET_FIRST(asp1);
      while(aip1)
      { if(!strcmp(s,"differ"))
        { if((asp2 = E_ASET_ITEM_VAL(aip1)) && E_IS_ASET(asp2))
          { if((aip2 = E_ASET_FIRST(asp2)))
            { mprintf("   sys    : ");
              print_file_props(E_ASET_ITEM_VAL(aip2),0);
            }
            if((aip2 = E_ASET_ITEM_NEXT(aip2)))
            { mprintf("   config : ");
              print_file_props(E_ASET_ITEM_VAL(aip2),0);
            }
            mprintf("\n");
          }
        }
        else
        { mprintf("    "); 
          print_file_props(E_ASET_ITEM_VAL(aip1),0);
        }
        aip1 = E_ASET_ITEM_NEXT(aip1);
      }
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }

  return(1);
}

/**************************************************************************/
/* Install the system files from the diffs aset:                          */
/* asp : aset of differences = {extra={}; missing={}; differ={{}{}}}      */
/* sys_file : system directory path name (prefix path for entry)          */
/* echo : if 1, echo each action.                                         */
/* mode : INSTALL_MODE_FIRST_INSTALL: If the r flag is set, the behaviour */ 
/*        is identical to install otherwise the only difference is that   */
/*        extra files are renamed with a .DIST extension rather than      */
/*        being removed. r flag check is made in the calling routine!     */
/* extra : entries are removed                                            */
/* missing : symbolic links and directories are created                   */
/*           other files and dirs are listed as being missing             */
/* differ : symbolic links are corrected                                  */
/*          files and directory attributes are corrected                  */
/**************************************************************************/
int dir_install(sysfile_info *sip,e_aset *asp,char *sys_file,int echo,int mode)
{ e_aset_item *aip,*aip1,*aip2;
  e_aset *asp1,*asp2,*asp_sys,*asp_conf;
  e_name *np;
  char *s=0;
  file_pars fp_sys,fp_conf;
  int matched=1;

  aip = E_ASET_FIRST(asp);
  while(aip)
  { if((np = E_ASET_ITEM_NAME(aip))) s = E_NAME_NAME(np);
    else s = 0;


    if((asp1 = E_ASET_ITEM_VAL(aip)) && (E_ASET_CARDINAL(asp1) > 0 ))
    { if(echo) mprintf("  %s :\n",s);
     
      if(!strcmp(s,"extra"))   /* remove the files that are extra */
      { aip1 = E_ASET_FIRST(asp1);
        while(aip1)
        { if(E_ASET_ITEM_VAL(aip1) && 
             fill_file_pars(E_ASET_ITEM_VAL(aip1),&fp_sys))
          { if((mode & INSTALL_MODE_FIRST_INSTALL) && 
               !(sip->flags & E_SYSF_FLAG_r))
            { matched=rename_dir_entry(sip,&fp_sys,sys_file,echo);
            }
            else
            { matched=remove_dir_entry(sip,&fp_sys,sys_file,echo);
            }
          }
          aip1 = E_ASET_ITEM_NEXT(aip1);
        }
      }
      else if(!strcmp(s,"missing"))   /* create the missing files */
      { aip1 = E_ASET_FIRST(asp1);
        while(aip1)
        { if(E_ASET_ITEM_VAL(aip1) && 
             fill_file_pars(E_ASET_ITEM_VAL(aip1),&fp_conf))
          { matched=create_dir_entry(sip,&fp_conf,sys_file,echo);
          }
          aip1 = E_ASET_ITEM_NEXT(aip1);
        }
      }
      else if(!strcmp(s,"differ"))    /* update files that differ */
      { aip1 = E_ASET_FIRST(asp1);
        while(aip1)
        { /* {{ sys }{ conf }} */
          if((asp2 = E_ASET_ITEM_VAL(aip1)) && (E_IS_ASET(asp2)))
          { if((aip2 = E_ASET_FIRST(asp2)))
            { asp_sys = E_ASET_ITEM_VAL(aip2);
              aip2 = E_ASET_ITEM_NEXT(aip2);
              asp_conf = E_ASET_ITEM_VAL(aip2);

              if(fill_file_pars(asp_sys,&fp_sys) &&
                 fill_file_pars(asp_conf,&fp_conf))
              { matched=install_dir_entry(sip,&fp_sys,&fp_conf,sys_file,
                                          echo,mode);
              }
            } 
          }
          aip1 = E_ASET_ITEM_NEXT(aip1);
        }
      }
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }
  return(matched);
}

/*************************************************************/
/* Remove the file described in the file_pars                */
/* { name; owner; group; modes; type; size; checksum; link;} */
/* Execute the auxiliary command with ACTION = REMOVE        */
/* fpp : system file parameters                              */
/* sys_file : path name for system the dir                   */
/* Return : 1 : the file has been removed                    */
/*          0 : The file has not been removed                */
/* Link : asp : 0                                            */
/*************************************************************/
int remove_dir_entry(sysfile_info *sip,file_pars *fpp,char *sys_file,int echo)
{ char Buff[E_BUFF_SIZE];

                /* Execute the auxiliary command */
  sprintf(Buff,"%s/%s",sys_file,fpp->name);
  execute_DL_aux_com(sip,Buff,fpp->type[0],"REMOVE",echo);

  if((fpp->type[0] == 'l') || (fpp->type[0] == 'f'))
  { /* remove the extra symbolic link or file */

    return(remove_file(Buff,echo));
  }
  
  mprintf("DIR LIST: file of type %s : %s/%s must be removed by hand!\n",
                 fpp->type,sys_file,fpp->name);
  
  return(0);
}

/*************************************************************/
/* Rename the file described in the file_pars by adding      */
/* the .DIST extension to the name. (regular files and       */
/* symbolic links only).                                     */
/* { name; owner; group; modes; type; size; checksum; link;} */
/* Execute the auxiliary command with ACTION = REMOVE        */
/* fpp : system file parameters                              */
/* sys_file : path name for system the dir                   */
/* Return : 1 : the file has been removed                    */
/*          0 : The file has not been removed                */
/* Link : asp : 0                                            */
/*************************************************************/
int rename_dir_entry(sysfile_info *sip,file_pars *fpp,char *sys_file,int echo)
{ char Buff[E_BUFF_SIZE],Buff2[E_BUFF_SIZE];
  struct stat stb;

          /* Execute the auxiliary command */
  sprintf(Buff,"%s/%s",sys_file,fpp->name);
  execute_DL_aux_com(sip,Buff,fpp->type[0],"REMOVE",echo);

  if((fpp->type[0] == 'l') || (fpp->type[0] == 'f'))
  {       /* remove the extra symbolic link or file */
    sprintf(Buff2,"%s/%s.DIST",sys_file,fpp->name);

    if(lstat(Buff2,&stb))
    { /* the .DIST file does not exist yet: OK */
      return(rename_file(Buff,Buff2,echo));
    }
    else
    { mprintf("DIR LIST: file %s already exists!\n",Buff2);
    }
  }
  else 
  { mprintf("DIR LIST: file of type %s : %s/%s must be removed by hand!\n",
                 fpp->type,sys_file,fpp->name);
  }

  return(0);
}

/*********************************************************/
/* Create the file described in the file par             */
/* fpp : config file parameters                          */
/* sys_file : path name for system the dir               */
/* echo : 1 => echo actions                              */
/* Return : 1 : the file has been removed                */
/*          0 : The file has not been created            */
/* Link : asp : 0                                        */
/*********************************************************/
int create_dir_entry(sysfile_info *sip,file_pars *fpp,char *sys_file,int echo)
{ char Buff[E_BUFF_SIZE];
  int matched=1;

  sprintf(Buff,"%s/%s",sys_file,fpp->name);

  if(fpp->type[0] == 'l')
  {            /* add the extra symbolic link */
    if(!make_sym_link(fpp->link,Buff,echo)) return(0);

               /* and set its ownership */
#ifdef ZZZXZ /* we don't set symbolic link ownership */
    if(lchown(Buff,fpp->owner,fpp->group))
    { mprintf("Unable to change the uid and gid for %s!",Buff);
      matched = 0;
    }
    else if(echo)
    { mprintf("Updating %s : uid=%d gid=%d\n",Buff,fpp->owner,fpp->group);
    }
#endif
  }
  else if(fpp->type[0] == 'd')
  {            /* create the missing directory */
    if(!mk_path_dir(Buff,echo)) return(0);

               /* and set its ownership */
    if(chown(Buff,fpp->owner,fpp->group))
    { mprintf("Unable to change the uid and gid for %s!",Buff);
      matched = 0;
    }
    else if(echo)
    { mprintf("Updating %s : uid=%d gid=%d\n",Buff,fpp->owner,fpp->group);
    }

                /* and set its modes. */
    if(chmod(Buff,fpp->modes))
    { mprintf("Unable to change the file mode for %s!",Buff);
      matched = 0;
    }
    else if (echo)
    { mprintf("Updating %s : mode=%o\n",Buff,fpp->modes);
    } 
  }
  else 
  { mprintf("DIR LIST: file of type %s : %s/%s must be added by hand!\n",
                fpp->type,sys_file,fpp->name);
    matched = 0;
  }

  execute_DL_aux_com(sip,Buff,fpp->type[0],"CREATE",echo);

  return(matched);
}

/*********************************************************/
/* Change a file from the system dir entry so that it    */
/* matches the config file                               */
/* fp_sys : system file parameters                       */
/* fp_conf : config file parameters                      */
/* sys_file : path name for system the dir               */
/* echo : 1 => echo actions                              */
/* Return : 1 : the file has been updated                */
/*          0 : The file has not been updated            */
/* Link : asp : 0                                        */
/*********************************************************/
int install_dir_entry(sysfile_info *sip,file_pars *fp_sys,file_pars *fp_conf,
                      char *sys_file,int echo,int mode)
{ char Buff[E_BUFF_SIZE];
  int matched=1;

  if((fp_sys->type[0] != fp_conf->type[0]) || (fp_sys->type[0] == 'l'))
  {      /* create the .DIST for symbolic links if necessary */
     if((mode & INSTALL_MODE_FIRST_INSTALL) && !(sip->flags & E_SYSF_FLAG_r))
     { if(rename_dir_entry(sip,fp_sys,sys_file,echo) &&
          create_dir_entry(sip,fp_conf,sys_file,echo))
       { return(1);
       }
     }
     else 
     {   /* files have different types or are symbolic links */
         /* remove old file and create the new one           */
  
       if(remove_dir_entry(sip,fp_sys,sys_file,echo) &&
          create_dir_entry(sip,fp_conf,sys_file,echo))
       { return(1);
       }
    }
    return(0);
  }

  sprintf(Buff,"%s/%s",sys_file,fp_sys->name);

        /* ownership */
  if ((fp_sys->owner != fp_conf->owner) || (fp_sys->group != fp_conf->group))
  { /* match uid and group */
    if(chown(Buff,fp_conf->owner,fp_conf->group))
    { mprintf("Unable to change the uid and gid for %s!",Buff);
      matched = 0;
    } 
    else if (echo)
    { mprintf("Updating %s : uid=%d gid=%d\n",
              Buff,fp_conf->owner,fp_conf->group);
    }   
  }

  if (fp_sys->modes != fp_conf->modes)
  { /* match mode except for symbolic links */
    if(chmod(Buff,fp_conf->modes))
    { mprintf("Unable to change the file mode for %s!",Buff);
      matched = 0;
    }
    else if (echo)
    { mprintf("Updating %s : mode=%o\n",Buff,fp_conf->modes);
    } 
  }

  if ((fp_sys->type[0] == 'f') && 
        ((fp_sys->size != fp_conf->size) || 
             strcmp(fp_sys->checksum,fp_conf->checksum))
     )
  { mprintf("DIR LIST: file of type %s : %s/%s must be updated by hand!\n",
                fp_sys->type,sys_file,fp_sys->name);
    matched = 0;
  }  

  return(matched);
}

/******************************************/
/* Compute the check sum for the file and */
/* return it as a string in buff          */
/* file_name : file to hash               */
/* buff : destination buffer for checksum */
/* return : 1 : OK                        */
/*          0 : error                     */
/******************************************/
int sysf_check_sum(char *file_name,char *buff,int echo)
{ unsigned int m;

  if (!strcmp(sysf_cf_vars.check_sum,"md5"))
  { if(!md5_hash(file_name,buff)) return(0);
  }
  else
  { if((m = hash_file(file_name))< 0) return(0);
    sprintf(buff,"%u",m);
  }


  return(1);
}

/****************************************/
/* compute a hash value for a file      */
/* file_name : name of the file to hash */
/* Return : hash code 0 <= HASH_Q2<     */
/*          0xffffffff : error          */
/****************************************/
unsigned int hash_file(char *file_name)
{ FILE *fpt;
  int c,i,size;
  unsigned int n,m,r,s;
  char buff[E_LONG_BUFF_SIZE];

  if(!(fpt = fopen(file_name,"r")))
  { printf("hash_file error : can't open file %s\n",file_name);
    return(0xffffffff);
  }

  m=r=1;

  while((size=fread(buff,1,E_LONG_BUFF_SIZE,fpt))>0)
  { i = 0;
    while(i < size)
    { c = buff[i++];
      if(c < 0) c = -c+128;

      if((c > 127) || ((c < 32) && (c != '\t') && (c!= '\n')  && (c!= '\r')))
      { s = r >> 1;
      }     
      else s = r;

      n = c+1;
      m = (((m*n)%HASH_Q1)+s)%HASH_Q1;
      r = ((m+1)*(r+1))%HASH_Q2;
    }
  }
  fclose(fpt);
  return(r);
}

/***********************************************/
/* compute the md5 hash value for a file       */
/* file_name : name of the file to hash        */
/* buff : buffer string result (char buff[40]) */
/* Return the md5 value in buff                */
/* Return : 1 : OK                             */
/*          0 : error                          */
/***********************************************/
unsigned int md5_hash(char *file_name,char *buff)
{ FILE *fpt;
  MD5_CTX context;
  int len;
  unsigned char buffer[E_LARGE_BUFF_SIZE], digest[16];
  unsigned int i,j;

  if(!(fpt = fopen(file_name,"r")))
  { printf("hash_file error : can't open file %s\n",file_name);
    return(0);
  }

  MD5Init (&context);
  while ((len = fread (buffer, 1, E_LARGE_BUFF_SIZE, fpt)))
  { MD5Update (&context, buffer, len);
  }
  MD5Final (digest, &context);

  fclose (fpt);

  j = 0;
  for (i = 0; i < 16; i++)
  { sprintf(&buff[j],"%02x", digest[i]);
    j += 2;
  }

  return(1);
}

/****************************************************************/
/* Execute the system command file for a DIR LIST if it exists. */
/* the command arguments are :                                  */
/* AUX_COM SYSTEM_FILE TYPE ACTION ECHO_ON                      */
/*   AUX_COM : command path :                                   */
/*   SYSTEM_FILE : full path for the modified system file.      */
/*   TYPE : file type : f,l,d, ...                              */
/*   ACTION : REMOVE or CREATE                                  */
/*   ECHO_ON : on or off                                        */
/*   SYSF_COM : sysf command name (install ...)                 */
/*                                                              */
/* The output is either displayed or emailed.                   */
/* Return 1 : OK                                                */
/*        0 : error                                             */
/****************************************************************/
int execute_DL_aux_com(sysfile_info *sip,char *sys_file, char type, 
                       char *action, int echo)
{ e_string *esp = 0;
  char *argv[7],*echo_s,type_s[2],ComName[E_LINE_SIZE];

  if(!sip || !sys_file || !action)
  { goto error;
  }

        /* No config file : no point to do anything */
  if(!sip->config_file.FileName[0] || !sip->aux_com.exists)
  { return(1);
  }

  get_command_name(ComName);

  type_s[0] = type;
  type_s[1] = 0;

  if(echo) echo_s = "on";
  else     echo_s = "off";

  argv[0] = sip->aux_com.FileName;
  argv[1] = sys_file;
  argv[2] = type_s;
  argv[3] = action;
  argv[4] = echo_s;
  argv[5] = ComName;
  argv[6] = 0;

  if((esp = e_exec_pipe_read(argv)))
  { mprintf("%s",E_STRING_S(esp));
    E_FREE(esp);
  }
  return(1);
  
error:
  if(esp) E_FREE(esp);
  return(0);
}
