#include "sysf.h"
#include "sysf_init.h"
#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>

extern sysf_cf sysf_cf_vars;

/***************************************************************/
/* Read the host.conf file into the list sysfile.hosts_groups. */
/* hosts_groups = { Group1 = {Host1 Host2 ... };               */
/*                  Group2 = {Host3 Host4 ... };               */
/* }                                                           */
/* config_dir : path for the configuration file.               */
/* host_file : name of the config file ["host.conf"]           */
/* rem_dead : 1 : remove members of the DEAD group.            */
/* The host names Hosti are strings                            */
/* Return : 1 : OK                                             */
/*          0 : error                                          */
/***************************************************************/
int read_sysf_hosts(char *config_dir,char *hosts_file)
{ e_tokeniser *tp = 0;
  e_aset *groups = 0,*grp = 0,*asp;
  e_name *group_name = 0;
  int state;
  e_flag mode = E_TOKENISER_MODE_HASH_COMMENT;
  e_string *file_path=0;

                  /* create host_file path name */
  E_NULL_ERROR(file_path = e_string_new(0,config_dir,E_PATH_SIZE));
  E_NULL_ERROR(file_path = e_string_append_buffer(file_path,"/")); 
  E_NULL_ERROR(file_path = e_string_append_buffer(file_path,hosts_file)); 

  if(!(tp = e_tokeniser_new(0,e_charstream_new(0,
                                E_CHARSTREAM_TYPE_FILE,
                                e_file_new(0,0,E_STRING_S(file_path),"r"),
                                E_MODE_READ),
                                mode,E_SYSF_DEF_WORD,E_SYSF_DEF_OP,
                            E_SYSF_DEF_PUNCT,E_SYSF_DEF_SPACES)))
  { goto error;
  }

  E_NULL_ERROR(groups = e_aset_new(0)); 

  state = STATE_NEW;
  while(e_tokeniser_next(tp) > 0)
  { switch(tp->type)
    { 
      case E_TOKENISER_WORD_TOKEN :
            if(state == STATE_NEW)         /* new groups */
            { if(!isupper((int)E_TOKENISER_BUFF(tp)[0]))
              { e_error_syntax("sysf: Syntax error in hosts.conf, line %d: \n      Group name (%s) must start with an upper case letter!\n",
                E_TOKENISER_LINE_NO(tp),E_TOKENISER_BUFF(tp));
                goto error;
	      }
              E_NULL_ERROR((group_name = e_name_new(E_TOKENISER_BUFF(tp))));

              state = STATE_EQUAL;
              E_NULL_ERROR(grp = e_aset_new(0)); 

	    }          
            else if( state == STATE_EQUAL) /* Error  */
            { goto error1;
	    }          
            else                         /* add 1 host to the group */
	    { if(isupper((int)E_TOKENISER_BUFF(tp)[0]))
	      {                          /* insert group */
                if(!(asp = (e_aset *)e_aset_get_value_by_key(groups,
					 e_name_new(E_TOKENISER_BUFF(tp)),0)))
                { e_error_syntax( 
   "sysf: Error in hosts.conf file line %d: undefined host group %s\n", 
                  E_TOKENISER_LINE_NO(tp),E_TOKENISER_BUFF(tp));
		  goto error;
                }
                E_NULL_ERROR(grp = e_aset_fill(grp,asp)); asp = 0;
	      }
              else
              { E_NULL_ERROR((grp = e_aset_add_item(grp,
                               e_name_new(E_TOKENISER_BUFF(tp)),
                               e_string_new(0,E_TOKENISER_BUFF(tp),0),
                               E_ASET_ADD_MODE_UNIQUE_KEY)));
	      }
	    }        
            break;

      case E_TOKENISER_OP_TOKEN :       /* read the = op */
            if((E_TOKENISER_BUFF(tp)[0] != '=') || (state != STATE_EQUAL))
               goto error1;

            state = STATE_HOST;
            break;

      case E_TOKENISER_PUNCT_TOKEN :       /* end of host list */
            if ((E_TOKENISER_BUFF(tp)[0] != ';') || (state != STATE_HOST))
               goto error1;
            
	                          /* Append host list to the group list */
            groups = e_aset_append_item(groups,group_name,grp);
            grp = 0; group_name = 0;

            state = STATE_NEW;

            break;
    }
  }

  if(tp) E_FREE(tp);
  if(file_path) E_FREE(file_path);

  sysfile.hosts_groups = groups;
  return(1);

error1:
   e_error_syntax(
   "sysf: Syntax error reading hosts.conf file line %d: unexpected char %c\n", 
                  E_TOKENISER_LINE_NO(tp),E_TOKENISER_BUFF(tp)[0]);
error:
  if(tp)         E_FREE(tp);
  if(grp)        E_FREE(grp);
  if(group_name) E_FREE(group_name);
  if(file_path)  E_FREE(file_path);
  return(0);
}

/*****************************************************************************/
/* Read the sysf.conf file into the list sysfile.configs.                    */
/* configs={ Conf1 = { Group1 System_file1 Config_name1 [Flag1]};            */
/*           Conf2 = { Group2 System_file2 Config_name2 [Flag2]};            */
/* }                                                                         */
/* Conf1 are sets of 3, 4 or 5 strings                                       */
/* config_dir : path for the configuration file.                             */
/* config_file : name of the config file ["sysf.conf"]                       */
/* Return : 1 : OK                                                           */
/*          0 : error                                                        */
/*****************************************************************************/
e_aset *read_sysf_config(char *config_dir,char *config_file)
{ e_tokeniser *tp = 0;
  e_aset *configs = 0,*inc_configs,*item_list = 0;
  e_name *item_name = 0;
  int state;
  e_flag mode = E_TOKENISER_MODE_HASH_COMMENT;
  e_string *file_path=0;

                  /* create host_file path name */
  E_NULL_ERROR(file_path = e_string_new(0,config_dir,E_PATH_SIZE));
  E_NULL_ERROR(file_path = e_string_append_buffer(file_path,"/")); 
  E_NULL_ERROR(file_path = e_string_append_buffer(file_path,config_file)); 

  if(!(tp = e_tokeniser_new(0,e_charstream_new(0,
                                 E_CHARSTREAM_TYPE_FILE,
                                 e_file_new(0,0,E_STRING_S(file_path),"r"),
                                 E_MODE_READ),
                            mode,E_SYSF_DEF_WORD,E_SYSF_DEF_OP,
                            E_SYSF_DEF_PUNCT,E_SYSF_DEF_SPACES)))
  { goto error;
  }

  E_NULL_ERROR(configs = e_aset_new(0)); 

  state = STATE_NEW;
  while(e_tokeniser_next(tp) > 0)
  { switch(tp->type)
    { 
      case E_TOKENISER_WORD_TOKEN :
            if(state == STATE_NEW)         /* new configs */
            { if(!strcmp(E_TOKENISER_BUFF(tp),"INCLUDE"))
              { state = STATE_INCLUDE;
                item_name = 0;
                item_list = 0;
              }
              else
              { E_NULL_ERROR((item_name = e_name_new(E_TOKENISER_BUFF(tp))));
                state = STATE_EQUAL;
                E_NULL_ERROR(item_list = e_aset_new(0)); 
              }
	    }          
            else if ((state == STATE_EQUAL)||(state == STATE_END)) /* Error  */
            { goto error1;
	    }          
            else if( state == STATE_INCLUDE) /* Read file  */
            { inc_configs = 
                 read_sysf_config(config_dir,E_TOKENISER_BUFF(tp));
              E_NULL_ERROR(configs = e_aset_append(configs,inc_configs));

              state = STATE_END;
	    }          
            else                         /* add 1 the config item */
	    { E_NULL_ERROR((item_list = e_aset_append_item(item_list,
                               e_name_new(E_TOKENISER_BUFF(tp)),
                               e_string_new(0,E_TOKENISER_BUFF(tp),0))));
	    }        
            break;

      case E_TOKENISER_OP_TOKEN :       /* read the = op */
           if((E_TOKENISER_BUFF(tp)[0] != '=') || (state != STATE_EQUAL))
               goto error1;

            state = STATE_ITEM;
            break;

      case E_TOKENISER_PUNCT_TOKEN :       /* end of host list */
            if ((E_TOKENISER_BUFF(tp)[0] != ';') || 
                ((state != STATE_ITEM) && (state != STATE_END)))
               goto error1;
	    if(item_name && item_list)
            { if(E_ASET_CARDINAL(item_list) < 3)
  	      { e_error_syntax(
             "sysf: Syntax error reading sysf.conf file line %d:\n      To few parameter for item %s\n",
                    E_TOKENISER_LINE_NO(tp),E_NAME_NAME(item_name));
                E_FREE(item_name);
                goto error;
	       }             

                       /* Append item list to the config list */
               configs = e_aset_append_item(configs,item_name,item_list);
            }

            item_list = 0; item_name = 0;

            state = STATE_NEW;

            break;
    }
  }

  if(tp) E_FREE(tp);
  if(file_path) E_FREE(file_path);

  return(configs);

error1:
   e_error_syntax(
    "sysf: Syntax error reading sysf.conf file line %d: unexpected char %c\n", 
                  E_TOKENISER_LINE_NO(tp),E_TOKENISER_BUFF(tp)[0]);
error:
  if(tp)        E_FREE(tp);
  if(item_list) E_FREE(item_list);
  if(item_name) E_FREE(item_name);
  if(file_path) E_FREE(file_path);
  return(0);
}

/***********************************************************/
/* Create a set with all the hosts selected in hosts list: */
/* a) Split the list into an aset.                         */
/* b) Resolve each group name                              */ 
/* Return : aset of host names.                            */
/*          0 : error                                      */
/* Link : Return +1                                        */
/***********************************************************/
e_aset *read_args_hosts(char *hosts_list,int flag)
{ e_aset *hosts = 0,*asp = 0,*asp2;
  e_string *sp;

  E_NULL_ERROR(asp = e_aset_split(e_string_new(0,hosts_list,0),0,","));
  E_NULL_ERROR(hosts = e_aset_new(0));

  e_aset_next(asp,E_NEXT_RESET,0);

  while(e_aset_next(asp,E_NEXT_GET_NEXT,(void **)&sp) > 0)
  { if(isupper((int)E_STRING_S(sp)[0]))
    {               /* Group : append list of hosts for that group */
      if(!(asp2 = (e_aset *)e_aset_get_value_by_key(sysfile.hosts_groups,
                                         e_name_new(E_STRING_S(sp)),0)))
      { e_reset();
        e_error_syntax(
    "sysf : invalid argument, Host Group \"%s\" not defined!",E_STRING_S(sp));
        goto error;
      }
      E_NULL_ERROR(hosts = e_aset_append(hosts,asp2)); 
    }
    else            /* 1 host : append the host name */
    { hosts = e_aset_append_item(hosts,e_name_new(E_STRING_S(sp)),sp);
      sp = 0;
      E_NULL_ERROR(hosts);
    }
    if(sp) E_FREE(sp);
  }

  E_FREE(asp);

  if(flag & FLAG_REM_DEAD)
  { hosts = remove_DEAD_hosts(hosts);
  }

  return(hosts);

error:
  if(asp) E_FREE(asp);
  if(hosts) E_FREE(hosts);
  return(0); 
 
}

/***********************************************************/
/* Remove the exclude_hosts from the host list (asp)       */
/* a) Split exclude_hosts into an aset.                    */
/* b) Remove each host from the list                       */ 
/*    hosts : initial host group.                          */
/*    exclude_hosts : list of hosts/GROUPS to exclude      */
/* Return : aset of host names.                            */
/*          0 : error                                      */
/* Link : hosts : -1                                       */
/*        Return +1                                        */
/***********************************************************/
e_aset *remove_args_hosts(e_aset *hosts,char *exclude_hosts)
{ e_aset *asp_eh=0,*asp,*asp2;
  e_string *sp;

  E_NULL_ERROR(asp = e_aset_split(e_string_new(0,exclude_hosts,0),0,","));
  E_NULL_ERROR(asp_eh = e_aset_new(0));

  e_aset_next(asp,E_NEXT_RESET,0);

  /* expand the list of hosts (groups -> list of hosts) */
  while(e_aset_next(asp,E_NEXT_GET_NEXT,(void **)&sp) > 0)
  { if(isupper((int)E_STRING_S(sp)[0]))
    {               /* Group : append list of hosts for that group */
      if(!(asp2 = (e_aset *)e_aset_get_value_by_key(sysfile.hosts_groups,
                                         e_name_new(E_STRING_S(sp)),0)))
      { e_reset();
        e_error_syntax(
    "sysf : invalid argument, Host Group \"%s\" not defined!",E_STRING_S(sp));
        goto error;
      }
      E_NULL_ERROR(asp_eh = e_aset_append(asp_eh,asp2));
      if(sp) E_FREE(sp);
    }
    else            /* 1 host : append the host name */
    { E_NULL_ERROR(asp_eh = 
         e_aset_append_item(asp_eh,e_name_new(E_STRING_S(sp)),sp));
    }
    sp = 0;
  }

  /* Prune the aset and make sure each key is unique */
  E_NULL_ERROR(hosts = e_aset_fill(e_aset_new(0),hosts));
  
  /* remove each host in asp_eh from hosts */
  E_NULL_ERROR(hosts = e_aset_diff_by_key(hosts,asp_eh,-1));

  E_FREE(asp);
  E_FREE(asp_eh);
 
  return(hosts);

error:
  if(asp) E_FREE(asp);
  if(asp_eh) E_FREE(asp_eh);
  if(hosts) E_FREE(hosts);
  return(0); 
 
}

/***********************************************************/
/* Remove any host in the DEAD group from the hosts aset   */
/* Return : aset of host names.                            */
/*          0 : error                                      */
/* Link : hosts : -1                                       */
/*        Return +1                                        */
/***********************************************************/
e_aset *remove_DEAD_hosts(e_aset *hosts)
{ e_aset *asp_dead;

  E_NULL_ERROR(hosts);

  if((asp_dead = 
     e_aset_get_value_by_key(sysfile.hosts_groups,e_name_new("DEAD"),0)))

  { E_NULL_ERROR(hosts = e_aset_diff_by_key(hosts,asp_dead,-1));
    E_FREE(asp_dead);
  }

  return(hosts);

error:
  if(hosts) E_FREE(hosts);
  return(0); 
 
}

/********************************************************************/
/* Build a proprety list of config groups (sysfile.configs).        */
/* The keys are the name of the groups and the value are the set of */
/* all the config file (as e_name) belonging to that group.         */
/* The set is saved in sysfile.config_groups.                       */
/* config_groups = { Group1 = { conf1 conf2 ... };                  */
/*                   Group2 = { conf3 conf4 ... };                  */
/* }                                                                */
/* Return : 1 OK                                                    */
/*          0 Error                                                 */
/********************************************************************/
int make_config_group_lists()
{ e_name *np,*group_name;
  e_aset *asp = 0,*config_groups = 0;
  e_string *group;

  e_aset_next(sysfile.configs,E_NEXT_RESET,0);
  
  E_NULL_ERROR(config_groups = e_aset_new(0));

             /* Scan the list of all config files         */
             /* Build the property list :                 */
             /* { grp1 = { conf1 conf2 ..}; grp2 = { ..}} */
  while(e_aset_next(sysfile.configs,E_NEXT_GET_CUR_KEY,(void **)&np) > 0)
  { e_aset_next(sysfile.configs,E_NEXT_GET_NEXT,(void **)&asp);

             /* The first item is the group name          */
    E_NULL_ERROR(group = e_aset_get_value_by_index(asp,CONFIG_GROUP_INDEX));

             /* Convert it to a  name                     */
    group_name = e_name_new(E_STRING_S(group));

             /* Property values are the config names      */
    E_NULL_ERROR(e_aset_add_property(config_groups,group_name,np,0));
  }

  sysfile.config_groups = config_groups;
  return(1);

error:
  if(asp) E_FREE(asp);
  if(config_groups) E_FREE(config_groups);
  return(0);
}


/****************************************************************/
/* Create an aset with all the host group this host belongs to  */
/* and return it as an aset.                                    */
/* The aset returned is an aset of named e_names. {name=name}   */
/* They are in the order in which the groups were defined.      */
/* Read the hosts groups from sysfile.hosts_groups              */
/* Return : 1 : OK                                              */
/*          0 : error                                           */
/* Link : host : 0                                              */
/*        RETURN : +1                                           */
/****************************************************************/
e_aset *get_this_host_groups(e_string *host)
{ e_aset *asp = 0,*h_grp = 0;
  e_name *np = 0;

  sysfile.this_host_groups = 0;
  E_NULL_ERROR(asp = e_aset_new(0));

  e_aset_next(sysfile.hosts_groups,E_NEXT_RESET,0);
  
             /* Scan all the host groups                  */
  while(e_aset_next(sysfile.hosts_groups,E_NEXT_GET_CUR_KEY,(void **)&np)>0)
  { e_aset_next(sysfile.hosts_groups,E_NEXT_GET_NEXT,(void **)&h_grp);

    if(e_aset_get_value_index(h_grp,E_USE(host),0) >= 0)
    {                    /* host belong to this group                   */
                         /* add the group name to the set as an e_name. */
      E_NULL_ERROR(asp = e_aset_append_item(asp,E_USE(np),E_USE(np)));
    }
    if(np) E_FREE(np); 
    np = 0;
  } 
  return(asp);

error:
  if(h_grp) E_FREE(h_grp);
  if(asp) E_FREE(asp);
  if(np) E_FREE(np);
  return(0);
}

/**************************************************************/
/* Build the aset of entries for the given system file        */
/* confs : config asset :                                     */ 
/* { entry_name1 = {Group system_file config_name [Flags] }   */
/*   entry_name2 = {Group system_file config_name [Flags] }   */
/* }                                                          */
/* RETURN : the aset of entry (e_name) as { name1,name2 }     */
/*          0 : not found                                     */
/* LINK : confs  : 0                                          */
/*        Return : +1                                         */
/**************************************************************/
e_aset *get_entry_name_for_system_file(e_aset *confs,char *str)
{ e_string *spt=0,*system_file;
  e_aset *asp,*entry_name_asp=0;
  e_aset_item *aip;
  e_name *entry_name=0;
  int n;

  E_NULL_ERROR(spt=e_string_new(0,str,0));

  E_NULL_ERROR(entry_name_asp = e_aset_new(0));

  aip = E_ASET_FIRST(confs);

  n = 0;
  while(aip)
  { entry_name = E_ASET_ITEM_NAME(aip);
    asp = E_ASET_ITEM_VAL(aip);
            /* The system file is the 2nd item (index 1) in the aset */
    if((system_file = e_aset_get_value_by_index(asp,1)))
    { if(E_EQUALS(spt,system_file))
      { entry_name_asp = e_aset_append_item(entry_name_asp,0,
                                            E_USE(entry_name));
        n++;
      }
      else E_FREE(system_file);
      system_file = 0;
    }

    aip = E_ASET_ITEM_NEXT(aip);
  }

  if(!n)
  { E_SAFE_FREE(entry_name_asp);
    entry_name_asp = 0;
  }

  E_SAFE_FREE(spt);
  return(entry_name_asp);

error:
  E_SAFE_FREE(spt);
  E_SAFE_FREE(entry_name_asp);
  return(0);
}

/************************* sysf.cf **********************/

/***************************************************************/
/* Read the /etc/sysf.cf file and return the value read on the */
/* sysf_cf structure.                                          */
/* The sysf_cf structure must be initialised to default values.*/
/* They are overwriten only for entries apearing in the file   */
/* SYNTAX :                                                    */
/*   config_dir = DIR_PATH                                     */
/*   config_file = FILE_NAME                                   */
/* Return : 1 : OK                                             */
/*          0 : error                                          */
/***************************************************************/
int read_etc_sysf_cf(sysf_cf *scp)
{ e_tokeniser *tp = 0;
  int state;
  e_flag mode = E_TOKENISER_MODE_HASH_COMMENT|E_TOKENISER_MODE_READ_STRING;
  char var_name[E_BUFF_SIZE],*s;

  if(!e_stat_file(SYSF_CF)) 
  { mprintf("WARNING %s does not exist!\n",SYSF_CF);
    return(1); /* /etc/sysf.cf does not exist */
  }
                  /* create host_file path name */
  if(!(tp = e_tokeniser_new(0,e_charstream_new(0,
                                E_CHARSTREAM_TYPE_FILE,
                                e_file_new(0,0,SYSF_CF,"r"),
                                E_MODE_READ),
                                mode,E_SYSF_DEF_WORD,E_SYSF_DEF_OP,
                            E_SYSF_DEF_PUNCT,E_SYSF_DEF_SPACES)))
  { goto error;
  }
  state = STATE_NEW;
  while(e_tokeniser_next(tp) > 0)
  { switch(tp->type)
    { case E_TOKENISER_STRING_TOKEN :
      case E_TOKENISER_WORD_TOKEN :
            if(state == STATE_NEW)         /* new variable */
            { strncpy(var_name,E_TOKENISER_BUFF(tp),E_BUFF_SIZE-1);
              var_name[E_BUFF_SIZE-1]=0;

              state = STATE_EQUAL;
	    }          
            else if (state == STATE_EQUAL) /* Error  */
            { goto error1;
	    }          
            else                         /* set var value */
	    { s = 0;

              if(!strcmp(var_name,"config_dir")) 
              { s = scp->config_dir; 
              }
              else if(!strcmp(var_name,"config_file"))
              { s = scp->config_file;
              }
              else if(!strcmp(var_name,"rsh_com"))
              { s = scp->rsh_com;
               }
              else if(!strcmp(var_name,"ssh_com"))
              { s = scp->ssh_com;
              }
              else if(!strcmp(var_name,"mail_com"))
              { s = scp->mail_com;
              }
              else if(!strcmp(var_name,"remote_shell"))
              { s = scp->remote_shell;
              }
              else if(!strcmp(var_name,"check_sum"))
              { s = scp->check_sum;
              }
              else
              { goto error2;
              }

              if(s)
              { strncpy(s,E_TOKENISER_BUFF(tp),E_BUFF_SIZE-1);
                s[E_BUFF_SIZE-1]=0;
              }

              state = STATE_END;
	    }        
            break;

      case E_TOKENISER_OP_TOKEN :       /* read the = op */
            if((E_TOKENISER_BUFF(tp)[0] != '=') || (state != STATE_EQUAL))
               goto error1;

            state = STATE_ITEM;
            break;

      case E_TOKENISER_PUNCT_TOKEN :       /* end of host list */
            if ((E_TOKENISER_BUFF(tp)[0] != ';') || (state != STATE_END))
               goto error1;
            
	                          /* Append host list to the group list */
            state = STATE_NEW;

            break;
    }
  }

  if(tp) E_FREE(tp);

  return(1);

error2:
   e_error_syntax(
   "sysf: Syntax error reading file %s line %d: invalid parameter name %s\n", 
                  SYSF_CF,E_TOKENISER_LINE_NO(tp),E_TOKENISER_BUFF(tp));
   goto error;

error1:
   e_error_syntax(
   "sysf: Syntax error reading file %s line %d: unexpected char %c\n", 
                  SYSF_CF,E_TOKENISER_LINE_NO(tp),E_TOKENISER_BUFF(tp)[0]);
error:
  if(tp)         E_FREE(tp);
  return(0);
}

/*****************************************************************/
/* Check that each parameter in the sysf_cf par has valid values */
/* scp : pointer to the sysf_cf parameters                       */
/* Return : 1 : OK                                               */
/*          0 : ERROR                                            */
/*****************************************************************/
int check_sysf_cf_pars(sysf_cf *scp)
{ int Err = 0,e,i,m;
  char Buff[2*E_PATH_SIZE];

  Err = check_sysf_cf_dir("config_dir",scp->config_dir,Err);

  i = e_tools_str_add(Buff,scp->config_dir,0,2*E_PATH_SIZE);
  i = e_tools_str_add(Buff,"/",i,2*E_PATH_SIZE);
  i = e_tools_str_add(Buff,scp->config_file,i,2*E_PATH_SIZE);

  /* check the config file in the config directory */
  if(!(m=e_stat_file(Buff)) || !S_ISREG(m))
  { if (!m) mprintf("sysf ERROR: %s %s does not exist in %s !\n",
            "config_file",scp->config_file,scp->config_dir);
    else    mprintf("sysf ERROR: %s %s is not a file in %s !\n",
            "config_file",scp->config_file,scp->config_dir);
    Err = 1;
  }

  e = Err;
  if (!strcmp(scp->remote_shell,"rsh"))
  { Err = check_sysf_cf_file("rsh_com",scp->rsh_com,Err);
  }
  else if (!strcmp(scp->remote_shell,"ssh")) 
  { Err = check_sysf_cf_file("ssh_com",scp->ssh_com,Err);
  }
  else if (strcmp(scp->remote_shell,"none")) 
  { Err = 1;
    mprintf("sysf ERROR: Invalid remote_shell : %s\n",scp->remote_shell);
  }
  Err = check_sysf_cf_file("mail_com",scp->mail_com,Err);

  if(Forgiving) /* ignore bad path */
  { Err = e;
  }

  if (strcmp(scp->check_sum,"md5") && strcmp(scp->check_sum,"sysf"))
  { mprintf("sysf ERROR: Invalid check_sum : %s\n",scp->check_sum);
    Err = 1;
  }

  return(!Err);
}

/*****************************************************/
/* Test if var is an existing directory              */
/* var : sysf_cf variable name (for error message)   */
/* path : directory path name                        */
/* current error status.                             */
/* Return : 0 : OK                                   */
/*          1 : error                                */
/*****************************************************/
int check_sysf_cf_dir(char *var,char *path,int err)
{ int m;

  if(!(m=e_stat_file(path)) || !S_ISDIR(m))
  { if (!m) mprintf("sysf ERROR: %s %s does not exist!\n",var,path);
    else    mprintf("sysf ERROR: %s %s is not a directory\n",var,path);
    err = 1;
  }

  return(err);
}

/*****************************************************/
/* Test if var is an existing file                   */
/* var : sysf_cf variable name (for error message)   */
/* path : file path name                             */
/* current error status.                             */
/* Return : 0 : OK                                   */
/*          1 : error                                */
/*****************************************************/
int check_sysf_cf_file(char *var,char *path,int err)
{ int m;

  if(!(m=e_stat_file(path)) || !S_ISREG(m))
  { if (!m) mprintf("sysf ERROR: %s %s does not exist!\n",var,path);
    else    mprintf("sysf ERROR: %s %s is not a file\n",var,path);
    err = 1;
  }

  return(err);
}

/*****************************************************/
/* Check that the config dir contains a subdirectory */
/* for every config group                            */
/* grp : aset of config groups                       */
/* Return : 1 : OK                                   */
/*          0 : error                                */
/* Link : grp : 0                                    */
/*****************************************************/
int check_config_groups(e_aset *grp,char *config_dir)
{ e_aset_item *asip;
  e_name *np;
  char Buff[E_LONG_BUFF_SIZE];
  int i,ret;

  asip = E_ASET_FIRST(grp);
  ret = 1;

  while(asip)
  { if((np = E_ASET_ITEM_NAME(asip)))
    { i = e_tools_str_add(Buff,config_dir,0,E_LONG_BUFF_SIZE);
      i = e_tools_str_add(Buff,"/",i,E_LONG_BUFF_SIZE);
      i = e_tools_str_add(Buff,E_NAME_NAME(np),i,E_LONG_BUFF_SIZE);
 
      if(!S_ISDIR(e_stat_file(Buff)))
      {  e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
     "ERROR : sysf : the config group directory \"%s\" does not exist!",Buff);
         ret = 0;
      }
    }
    asip = E_ASET_ITEM_NEXT(asip);
  }

  return(ret);
}

/*********************************************************/
/* Check that no config group has the name of entry name */
/* grp : aset of config groups { groupN = {} }           */
/* en_asp : aset of config entries { entry_nameN ={} }   */
/* Return : 1 : OK                                       */
/*          0 : ERROR                                    */
/* Link : grp : 0                                        */
/*        en_asp : 0                                     */
/*********************************************************/
int check_entries_and_config_groups(e_aset *grp,e_aset *en_asp)
{ e_aset *asp=0;
  e_aset_item *asip;
  e_name *np;

  if(!grp || !en_asp) return(0);
 
  E_NULL_ERROR(asp = e_aset_common_by_key(E_USE(en_asp),grp));

  if(E_ASET_CARDINAL(asp))
  { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
  "ERROR: sysf: the following config groups are used for and entry name.
Reconfigure the sysf.conf file!");

    asip = E_ASET_FIRST(asp);
 
    while(asip)
    { if((np = E_ASET_ITEM_NAME(asip)))
      { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
                      "  %s ",E_NAME_NAME(np)); 
      }
      asip = E_ASET_ITEM_NEXT(asip);
    }
    goto error;  
  }
  
  E_SAFE_FREE(asp);
  return(1);

error:  
  E_SAFE_FREE(asp);
  return(0);
}

/************************* MAIL *************************/
/**************************************************/
/* Print message in a string or on the screen    */
/* sysfile.message is null : print on the screen */
/* otherwise append to sysfile.message           */
/*************************************************/
void mprintf(const char *fmt, ...)
{ va_list args;
  char buff[E_LONG_BUFF_SIZE];

  va_start(args, fmt);
  if(sysfile.messages)
  { vsprintf(buff,fmt, args); 
    sysfile.messages = e_string_append_buffer(sysfile.messages,buff);
  }
  else
  { vprintf(fmt, args); 
  }
  va_end(args);
}

/*************************************************/
/* Email sysfile.message to address              */
/* Return : 1 : OK                               */
/*          0 : error                            */
/*************************************************/
int mail(char *address,char *subject)
{ FILE *pfpt;
  e_string *com;

  com = e_string_new(0,sysf_cf_vars.mail_com,E_LONG_BUFF_SIZE);

  com = e_string_append_buffer(com," -s \"");
  com = e_string_append_buffer(com,subject);
  com = e_string_append_buffer(com,"\" ");
  com = e_string_append_buffer(com,address);

  if(!(pfpt = (FILE *)popen(E_STRING_S(com),"w")))
  { mprintf("Error executing the command: %s\n",E_STRING_S(com));
    return(0);
  } 

  if(sysfile.messages) fputs(E_STRING_S(sysfile.messages),pfpt);

  pclose(pfpt);

  return(1);
}

/************************* TESTS *************************/

/*****************************/
/* Display the config groups */
/*****************************/
void display_config_groups()
{ e_name *np;

  e_aset_next(sysfile.config_groups,E_NEXT_RESET,0);

  while(e_aset_next(sysfile.config_groups,E_NEXT_GET_NEXT_KEY,(void **)&np)>0)
  { mprintf("%s ",E_NAME_NAME(np));
    E_FREE(np);
  }
  mprintf("\n");

}

/*************************************************/
/* Display all the config files by config groups */
/*************************************************/
void display_all_config_files()
{ e_string *sp;
  e_aset *asp;  
  e_name *np;

  e_aset_next(sysfile.config_groups,E_NEXT_RESET,0);

  while(e_aset_next(sysfile.config_groups,E_NEXT_GET_CUR_KEY,(void **)&np) > 0)
  { mprintf("%s =",E_NAME_NAME(np));

    if(e_aset_next(sysfile.config_groups,E_NEXT_GET_NEXT,(void **)&asp) > 0) 

    e_aset_next(asp,E_NEXT_RESET,0);
    while(e_aset_next(asp,E_NEXT_GET_NEXT,(void **)&sp) > 0)
    { mprintf(" %s",E_STRING_S(sp));
      E_FREE(sp);
    } 
    mprintf(";\n");
    E_FREE(asp);
  }
}

/*************************************************/
/* Display all the config files parameters       */
/*************************************************/
void display_all_configs()
{ e_string *sp;
  e_aset *asp;  
  e_name *np;

  e_aset_next(sysfile.configs,E_NEXT_RESET,0);

  while(e_aset_next(sysfile.configs,E_NEXT_GET_CUR_KEY,(void **)&np) > 0)
  { mprintf("%s =",E_NAME_NAME(np));

    if(e_aset_next(sysfile.configs,E_NEXT_GET_NEXT,(void **)&asp) > 0) 

    e_aset_next(asp,E_NEXT_RESET,0);
    while(e_aset_next(asp,E_NEXT_GET_NEXT,(void **)&sp) > 0)
    { mprintf(" %s",E_STRING_S(sp));
      E_FREE(sp);
    } 
    mprintf(";\n");
    E_FREE(asp);
  }
}

/*************************************************/
/* Display all the config files by config groups */
/* Return : 1 : OK                               */
/*          0 : error                            */
/*************************************************/
int display_one_group(char *group)
{ e_string *sp;
  e_aset *asp = 0;  
  e_name *np;

  E_NULL_ERROR(np = e_name_new(group));

  asp = e_aset_get_value_by_key(sysfile.config_groups,np,0);

  if(!asp)
  { mprintf("sysf : Unknown group : %s\n",group);
    goto error;
  }

  mprintf("%s =",group);

  e_aset_next(asp,E_NEXT_RESET,0);
  while(e_aset_next(asp,E_NEXT_GET_NEXT,(void **)&sp) > 0)
  { mprintf(" %s",E_STRING_S(sp));
    E_FREE(sp);
  } 
  mprintf(";\n");
  E_FREE(asp);
  
  return(1);

error:
  if(asp) E_FREE(asp);
  return(0);
}

/************************************/
/* Display the lists of host groups */
/************************************/
void display_hosts_groups()
{ e_string *sp;
  e_aset *asp;  
  e_name *np;

  e_aset_next(sysfile.hosts_groups,E_NEXT_RESET,0);

  while(e_aset_next(sysfile.hosts_groups,E_NEXT_GET_CUR_KEY,(void **)&np) > 0)
  { mprintf("%s =",E_NAME_NAME(np));

    if(e_aset_next(sysfile.hosts_groups,E_NEXT_GET_NEXT,(void **)&asp) > 0) 

    e_aset_next(asp,E_NEXT_RESET,0);
    while(e_aset_next(asp,E_NEXT_GET_NEXT,(void **)&sp) > 0)
    { mprintf(" %s",E_STRING_S(sp));
      E_FREE(sp);
    } 
    mprintf(";\n");
    E_FREE(asp);
    E_FREE(np);
  }
}

/****************************************************/
/* Display the lists of groups this host belongs to */
/* host : the host to search.                       */
/****************************************************/
void display_these_hosts_groups(e_aset *hosts)
{ e_string *sp=0;
  e_aset *asp=0;  
  e_name *np;

  E_NULL_ERROR(hosts);

  e_aset_next(hosts,E_NEXT_RESET,0);
  while(e_aset_next(hosts,E_NEXT_GET_NEXT,(void **)&sp) > 0)  
  { E_NULL_ERROR(asp = get_this_host_groups(sp));

    mprintf("%s =",E_STRING_S(sp));
    e_aset_next(asp,E_NEXT_RESET,0);
    while(e_aset_next(asp,E_NEXT_GET_NEXT_KEY,(void **)&np) > 0)
    { mprintf(" %s",E_NAME_NAME(np));
      E_FREE(np);
    }
    mprintf(";\n");
    E_FREE(asp); asp=0;
    E_FREE(sp); sp=0;
  }

  return;

error:
  if(asp)   E_FREE(asp);
  if(sp)    E_FREE(sp);
}

/************************************/
/* Display the lists of host groups */
/************************************/
void display_host_group(char *group)
{ e_string *sp;
  e_aset *asp;  
  e_name *np;

  e_aset_next(sysfile.hosts_groups,E_NEXT_RESET,0);

  while(e_aset_next(sysfile.hosts_groups,E_NEXT_GET_CUR_KEY,(void **)&np) > 0)
  { if(e_aset_next(sysfile.hosts_groups,E_NEXT_GET_NEXT,(void **)&asp) > 0) 

    if(!strcmp(group,E_NAME_NAME(np)))
    { e_aset_next(asp,E_NEXT_RESET,0);
      while(e_aset_next(asp,E_NEXT_GET_NEXT,(void **)&sp) > 0)
      { mprintf("%s ",E_STRING_S(sp));
        E_FREE(sp);
      } 
      E_FREE(asp);
    }
    E_FREE(np);
  }
  mprintf("\n");
}

/***************************************************************/
/* Display the lists of argument hosts : hosts - exclude_hosts */
/***************************************************************/
void display_arg_hosts()
{ e_name *np;
  int n;

  e_aset_next(sysfile.hosts,E_NEXT_RESET,0);

  n = 0;
  while(e_aset_next(sysfile.hosts,E_NEXT_GET_NEXT_KEY,(void **)&np) > 0)
  { if(n) mprintf(" %s",E_NAME_NAME(np));
    else mprintf("%s",E_NAME_NAME(np));

    E_FREE(np);
    n++;
  }
  mprintf("\n");
}

/***************************************************************/
/* Display the lists of argument entry names                   */
/***************************************************************/
void display_arg_entry_names(e_aset *asp)
{ e_name *np;
  int n;

  e_aset_next(asp,E_NEXT_RESET,0);

  n = 0;
  while(e_aset_next(asp,E_NEXT_GET_NEXT,(void **)&np) > 0)
  { if(n) mprintf(" %s",E_NAME_NAME(np));
    else mprintf("%s",E_NAME_NAME(np));

    E_FREE(np);
    n++;
  }
  mprintf("\n");
}

/***************************/
/* Display the config file */
/***************************/
void display_configs()
{ e_string *sp;

  if((sp = E_TO_STRING(sysfile.configs,0,0)))
  { mprintf("%s\n",E_STRING_S(sp));
    E_FREE(sp);
  }
}

/*****************************/
/* Display the config groups */
/*****************************/
void find_group()
{ e_name *np,*pname;
  e_aset *asp;
  e_aset_item *aip;

  e_aset_next(sysfile.arg_configs,E_NEXT_RESET,0);

  while(e_aset_next(sysfile.arg_configs,E_NEXT_GET_NEXT,(void **)&np)>0)
  { if(np && E_IS_NAME(np))
    { mprintf("%s :",E_NAME_NAME(np));
    
      E_NULL_ERROR(asp=e_aset_get_properties(E_USE(sysfile.config_groups),
                    E_USE(np)));

      aip = E_ASET_FIRST(asp);
      while(aip)
      { if((pname = E_ASET_ITEM_NAME(aip)))
        { mprintf(" %s",E_NAME_NAME(pname));
        }
        aip = E_ASET_ITEM_NEXT(aip);
      }
      E_FREE(asp);
      mprintf("\n");
    }
  }
  return;

error:
  E_SAFE_FREE(asp);
  return;
}

/***************************************************************/ 
/* Make the entry name aset from an aset of strings            */
/* (entry names and config groups)                             */
/* asp : asp of strings : entry names and config groups        */
/* arg_name : "entry names" or "-exclude_entry_names"          */
/* Return : aset of entry names from arg                       */
/*          0 : ERROR                                          */
/* Link : asp : -1                                             */
/*        Return : +1                                          */
/***************************************************************/ 
e_aset *make_entry_name_aset(e_aset *asp,char *arg_name)
{ e_aset *enasp=0,*grp;
  e_aset_item *asip;
  e_name *np,*cnp;
  e_string *stp;

  E_NULL_ERROR(asp);
  E_NULL_ERROR(enasp=e_aset_new(0));

  asip = E_ASET_FIRST(asp);

  while(asip)
  { E_NULL_ERROR(stp = E_ASET_ITEM_VAL(asip));
    E_NULL_ERROR(np = e_name_new(E_STRING_S(stp)));

                        /* A config file */
    if((stp = e_aset_get_value_by_key(sysfile.configs,E_USE(np),0)))
    {                  /* Add the name to arg_configs */
      E_NULL_ERROR(enasp = e_aset_add_item(enasp,0,np,
                                           E_ASET_ADD_MODE_UNIQUE_VAL));
      np = 0;
      E_FREE(stp); stp = 0;
    }
                      /* A config group */
    else if((grp=e_aset_get_value_by_key(sysfile.config_groups,
                                         E_USE(np),0))) 
    {               /* Add the set of names to arg_configs */
      E_NULL_ERROR(enasp = 
         e_aset_add_aset(enasp,grp,E_ASET_ADD_MODE_UNIQUE_VAL));
      grp = 0;
    }
                        /* ALL : All config files */
    else if(!strcmp("ALL",E_NAME_NAME(np))) 
    {               /* Add the set of names to arg_configs */
      e_aset_next(sysfile.configs,E_NEXT_RESET,0);
  
                      /* Scan the list of all config files         */
      while(e_aset_next(sysfile.configs,E_NEXT_GET_CUR_KEY,
                        (void **)&cnp)>0)
      { e_aset_next(sysfile.configs,E_NEXT_GET_NEXT,(void **)&asp);
        if(asp) { E_FREE(asp); asp = 0;}

                        /* Property values are the config names      */
         E_NULL_ERROR(enasp = e_aset_add_item(enasp,0,
                                              cnp,E_ASET_ADD_MODE_UNIQUE_VAL));
         cnp = 0;
      }
    }
    else
    { e_error(E_ERROR_BAD_ARGUMENT,E_ERROR_LEVEL_USER,
                  "sysf error: %s list, invalid config file or group %s\n",
                  arg_name,
                  E_STRING_S(E_ASET_ITEM_VAL(asip)));
      goto error;
    }
    E_SAFE_FREE(np);
    np = 0;
    asip = E_ASET_ITEM_NEXT(asip);
  }

  E_SAFE_FREE(asp);
  return(enasp);

error:
  E_SAFE_FREE(enasp);
  E_SAFE_FREE(asp);
  return(0);
}
